from pathlib import Path
import os
import logging
from copy import copy

from qtpy import QtCore
from qtpy.QtCore import Qt
import numpy as np

from refnx._lib import preserve_cwd
from refnx.dataset import ReflectDataset
from refnx.reduce import (
    number_datafile,
    PlatypusReduce,
    basename_datafile,
    PlatypusNexus,
)


reducer_entry = [
    ("use", False),
    ("scale", 1.0),
    ("reflect-1", ""),
    ("reflect-2", ""),
    ("reflect-3", ""),
    ("direct-1", ""),
    ("direct-2", ""),
    ("direct-3", ""),
    ("flood", ""),
]
default_reducer_entry = {re[0]: re[1] for re in reducer_entry}


class ReductionState:
    """
    Reduces multiple reflectometry datafiles. Attributes control how the
    reduction proceeds.
    """

    def __init__(self, manual_beam_finder=None):
        super().__init__()

        # Each entry specifies the scale factor, whether the file is used for
        # reduction and the reflected and direct beam runs
        self.reduction_entries = {}

        self.default_reduction_options = {
            "low_wavelength": (2.5, float),
            "high_wavelength": (19, float),
            "rebin_percent": (2.0, float),
            "expected_centre": (500.0, float),
            "manual_beam_find": (False, bool),
            "background_subtraction": (True, bool),
            "monitor_normalisation": (True, bool),
            "save_offspecular": (True, bool),
            "save_spectrum": (True, bool),
            "streamed_reduction": (False, bool),
        }

        # the time slices for streamed reduction
        self.stream_start = 0
        self.stream_end = 3600
        self.stream_duration = 30

        # where the data, streamed data and output files are located
        self.data_directory = ""
        self.streamed_directory = ""
        self.output_directory = ""

        self.manual_beam_finder = manual_beam_finder

        for attr, val in self.default_reduction_options.items():
            setattr(self, attr, val[0])

        self.peak_pos = (90, 5)

        self.save_state_path = None

    def __getstate__(self):
        d = self.__dict__
        e = copy(d)
        e.pop("manual_beam_finder")
        return e

    @preserve_cwd
    def reducer(self, callback=None):
        """
        Reduce all the entries in reduction_entries

        Parameters
        ----------
        callback : callable
            Function, `f(percent_finished)` that is called with the current
            percentage progress of the reduction
        """

        # refnx.reduce.reduce needs you to be in the directory where you're
        # going to write files to
        if self.output_directory:
            os.chdir(self.output_directory)

        # if no data directory was specified then assume it's the cwd
        data_directory = self.data_directory
        if not data_directory:
            data_directory = "./"

        def full_path(fname):
            f = Path(data_directory) / fname
            return f

        # if the streamed directory isn't mentioned then assume it's the same
        # as the data directory
        streamed_directory = self.streamed_directory
        if not Path(streamed_directory).is_dir():
            self.streamed_directory = data_directory

        logging.info(
            "-------------------------------------------------------"
            "\nStarting reduction run"
        )
        logging.info(
            "data_folder={data_directory}, trim_trailing=True, "
            "lo_wavelength={low_wavelength}, "
            "hi_wavelength={high_wavelength}, "
            "rebin_percent={rebin_percent}, "
            "normalise={monitor_normalisation}, "
            "background={background_subtraction} "
            "eventmode={streamed_reduction} "
            "event_folder={streamed_directory}".format(**self.__dict__)
        )

        # sets up time slices for event reduction
        if self.streamed_reduction:
            eventmode = np.arange(
                self.stream_start, self.stream_end, self.stream_duration
            )
            eventmode = np.r_[eventmode, self.stream_end]
        else:
            eventmode = None

        # are you manual beamfinding?
        peak_pos = None
        if self.manual_beam_find and self.manual_beam_finder is not None:
            peak_pos = -1

        idx = 0

        cached_direct_beams = {}

        for row, val in self.reduction_entries.items():
            if not val["use"]:
                continue

            flood = None
            if val["flood"]:
                flood = full_path(val["flood"])

            combined_dataset = None

            # process entries one by one
            for ref, db in zip(
                ["reflect-1", "reflect-2", "reflect-3"],
                ["direct-1", "direct-2", "direct-3"],
            ):
                reflect = val[ref]
                direct = val[db]

                # if the file doesn't exist there's no point continuing
                if (not full_path(reflect).is_file()) or (
                    not full_path(direct).is_file()
                ):
                    continue

                # which of the nspectra to reduce (or all)
                ref_pn = PlatypusNexus(full_path(reflect))

                if direct not in cached_direct_beams:
                    cached_direct_beams[direct] = PlatypusReduce(
                        direct, data_folder=data_directory
                    )

                reducer = cached_direct_beams[direct]

                try:
                    reduced = reducer(
                        ref_pn,
                        scale=val["scale"],
                        h5norm=flood,
                        lo_wavelength=self.low_wavelength,
                        hi_wavelength=self.high_wavelength,
                        rebin_percent=self.rebin_percent,
                        normalise=self.monitor_normalisation,
                        background=self.background_subtraction,
                        manual_beam_find=self.manual_beam_finder,
                        peak_pos=peak_pos,
                        eventmode=eventmode,
                        event_folder=streamed_directory,
                    )
                except Exception as e:
                    # typical Exception would be ValueError for non overlapping
                    # angles
                    logging.info(e)
                    continue

                logging.info(
                    "Reduced {} vs {}, scale={}, angle={}".format(
                        reflect,
                        direct,
                        val["scale"],
                        reduced[1]["omega"][0, 0],
                    )
                )

                if combined_dataset is None:
                    combined_dataset = ReflectDataset()

                    fname = basename_datafile(reflect)
                    fname_dat = Path(self.output_directory) / f"c_{fname}.dat"
                    fname_xml = Path(self.output_directory) / f"c_{fname}.xml"

                try:
                    combined_dataset.add_data(
                        reducer.data(),
                        requires_splice=True,
                        trim_trailing=True,
                    )
                except ValueError as e:
                    # datasets don't overlap
                    logging.info(e)
                    continue

            if combined_dataset is not None:
                # after you've finished reducing write a combined file.
                with open(fname_dat, "wb") as f:
                    combined_dataset.save(f)
                with open(fname_xml, "wb") as f:
                    combined_dataset.save_xml(f)
                logging.info(
                    "Written combined files: {} and {}".format(
                        fname_dat, fname_xml
                    )
                )

            # can be used to create a progress bar
            idx += 1
            if callback is not None:
                ok = callback(100 * idx / len(self.reduction_entries))
                if not ok:
                    break

        logging.info(
            "\nFinished reduction run"
            "-------------------------------------------------------"
        )


class ReductionTableModel(QtCore.QAbstractTableModel):
    """
    a model for displaying in a QtGui.QTableView
    """

    def __init__(self, reduction_state, parent=None):
        super().__init__(parent)
        self._reduction_state = reduction_state

    @property
    def reduction_state(self):
        return self._reduction_state

    def rowCount(self, parent=QtCore.QModelIndex()):
        return 200

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(reducer_entry)

    def headerData(
        self, section, orientation, role=Qt.ItemDataRole.DisplayRole
    ):
        """Set the headers to be displayed."""
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if orientation == Qt.Orientation.Vertical:
            return None

        if orientation == Qt.Orientation.Horizontal:
            return reducer_entry[section][0]

        return None

    def flags(self, index):
        # row = index.row()
        col = index.column()
        if not col:
            return Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled

        return (
            Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
        )

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return False

        row = index.row()
        col = index.column()

        state = self.reduction_state
        attr_name = reducer_entry[col][0]

        if row in state.reduction_entries:
            entry = state.reduction_entries[row]
        else:
            entry = default_reducer_entry

        value = entry[attr_name]

        if role == Qt.ItemDataRole.CheckStateRole:
            if not col and value:
                return Qt.CheckState.Checked
            elif not col:
                return Qt.CheckState.Unchecked

        if role == Qt.ItemDataRole.DisplayRole and col:
            if attr_name == "scale":
                return str(float(value))
            else:
                return value

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        row = index.row()
        col = index.column()

        if not index.isValid():
            return False

        state = self.reduction_state
        attr_name = reducer_entry[col][0]

        if row in state.reduction_entries:
            entry = state.reduction_entries[row]
        else:
            entry = {re[0]: re[1] for re in reducer_entry}
            state.reduction_entries[row] = entry

        if role == Qt.ItemDataRole.CheckStateRole and col == 0:
            entry["use"] = value == Qt.CheckState.Checked

        if role == Qt.ItemDataRole.EditRole:
            if col == 0:
                save_value = False
                if value == Qt.CheckState.Checked:
                    save_value = True
            elif col == 1:
                try:
                    save_value = float(value)
                except ValueError:
                    save_value = 1
            else:
                # you're editing reflect/direct beam names
                try:
                    save_value = int(value)
                    save_value = number_datafile(save_value)
                except ValueError:
                    if value:
                        if value.endswith(".nx.hdf"):
                            save_value = value
                        else:
                            save_value = value + ".nx.hdf"
                    else:
                        return False

            entry[attr_name] = save_value

        self.dataChanged.emit(index, index)
        return True



"""
*Calculates the specular (Neutron or X-ray) reflectivity from a stratified
series of layers.

The refnx code is distributed under the following license:

Copyright (c) 2025 A. R. J. Nelson, ANSTO

Permission to use and redistribute the source code or binary forms of this
software and its documentation, with or without modification is hereby
granted provided that the above notice of copyright, these terms of use,
and the disclaimer of warranty below appear in the source code and
documentation, and that none of the names of above institutions or
authors appear in advertising or endorsement of works derived from this
software without specific prior written permission from all parties.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THIS SOFTWARE.

"""

from importlib import resources
import numpy as np
import numpy.typing as npt
from typing import Optional

from refnx.analysis import possibly_create_parameter, Parameter, Parameters
from refnx.reflect.reflect_model import (
    ReflectModel,
    SpinChannel,
    kernel,
    reflectivity,
)

try:
    from refnx.reflect._creflect import gepore
except ImportError:

    def gepore(*args, **kwds):
        raise RuntimeError("gepore is not available")


class _MemoisingReflectivityKernel:
    def __init__(self, Aguide=270):
        self._memoised_x = None
        self._memoised_slabs = None
        self._memoised_reflectivity = None
        self.updated = False
        self.c = {
            SpinChannel.UP_UP: 0,
            SpinChannel.UP_DOWN: 1,
            SpinChannel.DOWN_UP: 2,
            SpinChannel.DOWN_DOWN: 3,
        }
        self.spin = SpinChannel.UP_UP
        self.Aguide = Aguide

    def __call__(self, x, slabs, *args, **kwds):
        shp = x.shape

        # calculate all four spin channels for a given set of x.
        if (
            self.updated
            and np.array_equal(self._memoised_x, x)
            and np.array_equal(self._memoised_slabs, slabs)
        ):
            return self._memoised_reflectivity[self.c[self.spin]].reshape(shp)

        # if x is being called by a smearing kernel it may be presented in a
        # N-dimensional fashion, e.g. with shape (N, M).
        # gepore will always return an array of shape (N*M, 4). We therefore
        # store the memoised reflectivity in a flattened fashion, but reshape
        # the correct spin state to (N, M) on return.
        xflat = np.ravel(x)
        kwds["Aguide"] = self.Aguide
        R = gepore(xflat, slabs, *args, **kwds)
        self._memoised_reflectivity = R
        self._memoised_slabs = slabs
        self._memoised_x = x
        self.updated = True
        return R[self.c[self.spin]].reshape(shp)


class PolarisedReflectModel(ReflectModel):
    """
    Extension of ReflectModel for polarised neutron reflectometry.
    See `refnx.reflect.ReflectModel` for documentation of arguments.

    Parameters
    ----------
    structure : refnx.reflect.Structure
        The interfacial structure.
    scales : {float, refnx.analysis.Parameter, sequence, refnx.analysis.Parameters}, optional
        Scale factor for each spin channel. All model values are multiplied by
        this(/these) value(s) before the background is added. They are turned into a
        Parameter during the construction of this object. If an iterable is
        provided it should contain four values (one for each channel), in the
        order:

            - SpinChannel.UP_UP
            - SpinChannel.UP_DOWN
            - SpinChannel.DOWN_UP
            - SpinChannel.DOWN_DOWN

    bkgs : {float, refnx.analysis.Parameter, sequence, refnx.analysis.Parameters}, optional
        Background for each spin channel. Q-independent constant background
        added to all model values. They are turned into a Parameter during
        the construction of this object. If an iterable is
        provided it should contain four values (one for each channel).
    name : str, optional
        Name of the Model
    q_offsets: {float, refnx.analysis.Parameter, sequence, refnx.analysis.Parameters}, optional
        Compensates for uncertainties in the angle at which the measurement is
        performed. A positive/negative `q_offset` corresponds to a situation
        where the measured q values (incident angle) may have been under/over
        estimated, and has the effect of shifting the calculated model to
        lower/higher effective q values. They are turned into a Parameter during
        the construction of this object. If an iterable is
        provided it should contain four values (one for each channel).
    Aguide: float
        Angle of applied field. This value should be 270 or 90 degrees for
        the applied field to lie in the plane of the sample, perpendicular to
        the beam propagation direction. For a magnetic moment to be parallel
        or anti-parallel to the applied field `thetaM` should be 90 or -90 deg
        respectively.
    """

    def __init__(
        self,
        structure,
        scales=1.0,
        bkgs=1e-7,
        name="",
        dq=5.0,
        threads=-1,
        quad_order=17,
        dq_type="constant",
        q_offsets=0,
        Aguide=270,
    ):
        super().__init__(
            structure,
            name=name,
            threads=threads,
            quad_order=quad_order,
            dq=dq,
            dq_type=dq_type,
        )
        # overwrite the scales/bkgs/q_offsets
        if hasattr(scales, "__iter__"):
            _s = [possibly_create_parameter(v) for v in scales]
            if len(_s) != 4:
                raise RuntimeError(
                    "Four scale values need to be provided, one for each spin channel"
                )
            self._scale = Parameters(name="scales", data=_s)
        else:
            self._scale = Parameter(scales, name="scale")

        if hasattr(bkgs, "__iter__"):
            _s = [possibly_create_parameter(v) for v in bkgs]
            if len(_s) != 4:
                raise RuntimeError(
                    "Four bkg values need to be provided, one for each spin channel"
                )
            self._bkg = Parameters(name="bkgs", data=_s)
        else:
            self._bkg = Parameter(bkgs, name="bkg")

        if hasattr(q_offsets, "__iter__"):
            _s = [possibly_create_parameter(v) for v in q_offsets]
            if len(_s) != 4:
                raise RuntimeError(
                    "Four q_offset values need to be provided, one for each spin channel"
                )
            self._q_offset = Parameters(name="q_offsets", data=_s)
        else:
            self._q_offset = Parameter(q_offsets, name="q_offset")

        # update internal parameter view
        self.structure = structure
        self.Aguide = Aguide
        self._memoising_kernel = _MemoisingReflectivityKernel()

    def __repr__(self):
        return (
            f"PolarisedReflectModel({self._structure!r}, name={self.name!r},"
            f" scales={self.scale!r}, bkgs={self.bkg!r},"
            f" dq={self.dq!r}, threads={self.threads},"
            f" quad_order={self.quad_order!r}, dq_type={self.dq_type!r},"
            f" q_offsets={self.q_offset!r}, Aguide={self.Aguide!r})"
        )

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, val):
        raise RuntimeError("Setter disabled in subclass")

    @property
    def bkg(self):
        return self._bkg

    @bkg.setter
    def bkg(self, val):
        raise RuntimeError("Setter disabled in subclass")

    @property
    def q_offset(self):
        return self._q_offset

    @q_offset.setter
    def q_offset(self, val):
        raise RuntimeError("Setter disabled in subclass")

    def model(self, x, p=None, x_err=None, spin="all"):
        r"""
        Calculate the reflectivity of this model

        Parameters
        ----------
        x : float or np.ndarray
            q values for the calculation. This should be a 2-D array with
            4 columns.
            In order the 4 columns represent:

                - SpinChannel.UP_UP
                - SpinChannel.UP_DOWN
                - SpinChannel.DOWN_UP
                - SpinChannel.DOWN_DOWN

            In a given row only one column should contain a finite value, the
            others should be set to np.nan.
            If a single spin channel is requested then `x` can be 1-D.
            Units = Angstrom**-1
        p : refnx.analysis.Parameters, optional
            parameters required to calculate the model
        x_err : {np.ndarray, float} optional
            Specifies how the instrumental resolution smearing is carried out
            for each of the points in `x`. If an array it should have the same
            shape as `x`.
            See :func:`refnx.reflect.reflectivity` for further details.
        spin : {"all", SpinChannel}, optional
            Specifies which spin to return from the calculation. If all then
            `x` needs to be 2-D. If a single spin channel is requested then
            `x` can be 1-D.

        Returns
        -------
        reflectivity : np.ndarray
            Calculated reflectivity

        Notes
        -----
        If `x_err` is not provided then the calculation will fall back to
        the constant dq/q smearing specified by the `dq` attribute of this
        object.
        """
        if p is not None:
            self.parameters.pvals = np.array(p)
        if x_err is None or self.dq_type == "constant":
            # fallback to what this object was constructed with
            x_err = float(self.dq)

        if spin != "all":
            # only want to calculate one spin channel
            pass
        elif (
            x.ndim != 2
            or x.shape[1] != 4
            or np.any(np.count_nonzero(np.isfinite(x), axis=1) > 1)
        ):
            raise RuntimeError(
                "For PolarisedReflectModel.model the x array should have shape"
                " (N, 4). Each of the columns represents a SpinChannel. Only"
                " one of the entries in a row should be finite, the rest"
                " should be np.nan"
            )

        slabs = self.structure.slabs()
        if slabs.shape[1] == 5:
            # unpolarised for some reason
            raise ValueError(
                "Please use ReflectModel, not PolarisedReflectModel,"
                " for non-magnetic systems."
            )
            # slabs = slabs[..., :4]
            # return reflectivity(
            #     _x_union,
            #     slabs,
            #     scale=self.scale[0].value,
            #     bkg=self.bkg[0].value,
            #     dq=x_err,
            #     threads=self.threads,
            #     quad_order=self.quad_order,
            #     q_offset=self.q_offset[0].value,
            # )
        if slabs.shape[1] == 7:
            # polarised
            slabs = np.take_along_axis(
                slabs, np.array([0, 1, 2, 3, 5, 6])[None, :], axis=1
            )

        if isinstance(self.scale, Parameters):
            scale = np.array(self.scale)
        else:
            scale = [self.scale.value] * 4

        if isinstance(self.bkg, Parameters):
            bkg = np.array(self.bkg)
        else:
            bkg = [self.bkg.value] * 4

        if spin != "all":
            # calculate single spin channel
            self._memoising_kernel.spin = spin
            idx = self._memoising_kernel.c[spin]
            scale = scale[idx]
            bkg = bkg[idx]
            _R = reflectivity(
                x,
                slabs,
                scale=scale,
                bkg=bkg,
                dq=x_err,
                threads=self.threads,
                quad_order=self.quad_order,
                fkernel=self._memoising_kernel,
            )
            return _R

        # calculate all spin channels at once
        _x_union = self._calculate_x_union(x)
        _x_err_union = self._calculate_x_err_union(x, x_err)
        active_spins = self._active_spins(x)

        # if we're only calculating NSF, there's no need to use gepore
        # shortcut to use unpolarised calculator.
        def is_there_spin_flip(slabs):
            return any(
                np.logical_and(
                    slabs[:, 4] != 0.0,
                    np.abs(np.sin(np.radians(slabs[:, 5]))) < 1,
                )
            )

        if (
            not any(active_spins[1:3])  # UU, UD, DU, DD
            and self.Aguide == 270
            and not is_there_spin_flip(slabs)
        ):
            return self._calculate_nsf_only(
                active_spins, slabs, x, x_err, scale, bkg
            )

        # calculate smeared reflectivity on the entire _x_union
        R_union = np.zeros(len(x))

        for idx, spin in enumerate(SpinChannel):
            if not active_spins[idx]:
                continue

            msk = np.isfinite(x[:, idx])
            self._memoising_kernel.spin = spin

            _R = reflectivity(
                _x_union,
                slabs,
                dq=_x_err_union,
                threads=self.threads,
                quad_order=self.quad_order,
                fkernel=self._memoising_kernel,
            )
            R_union[msk] = _R[msk] * scale[idx] + bkg[idx]

        return R_union

    def _calculate_x_union(self, x):
        if isinstance(self.q_offset, Parameters):
            q_offsets = np.array(self.q_offset)
        else:
            q_offsets = [self.q_offset.value] * 4

        _x = x.copy()
        for idx in range(4):
            _x[:, idx] += q_offsets[idx]

        return np.nanmax(_x, axis=1)

    def _calculate_x_err_union(self, x, x_err):
        if isinstance(x_err, np.ndarray):
            if x.shape != x_err.shape:
                raise RuntimeError(
                    f"x_err.shape and x.shape should be equal"
                    f" {x_err.shape=}, {x.shape=}"
                )
            x_err_union = np.zeros(len(x))
            for idx in range(4):
                msk = np.isfinite(x[:, idx])
                x_err_union[msk] = x_err[msk, idx]

            return x_err_union
        else:
            # single number represents constant dq/q
            return x_err

    def _active_spins(self, x):
        return np.any(np.isfinite(x), axis=0)

    def _calculate_nsf_only(self, active_spins, slabs, x, x_err, scale, bkg):
        """
        Calculates NSF reflectivities only.

        Parameters
        ----------
        active_spins : np.ndarray
            Length-4 boolean array specifying which spin channels are being
            requested. UU, UD, DU, DD
        slabs : np.ndarray
            Slab representation of the interfacial model.
            See `refnx.reflect.reflectivity` for further details
        x : np.ndarray
            q values for the calculation. This should be a 2-D array with
            4 columns.
            In order the 4 columns represent:

                - SpinChannel.UP_UP
                - SpinChannel.UP_DOWN
                - SpinChannel.DOWN_UP
                - SpinChannel.DOWN_DOWN

            In a given row only one column should contain a finite value, the
            others should be set to np.nan. For this NSF function there should
            be no finite values in columns 1, 2.
            Units = Angstrom**-1
        x_err : {np.ndarray, float} optional
            Specifies how the instrumental resolution smearing is carried out
            for each of the points in `x`.
            See :func:`refnx.reflect.reflectivity` for further details.
        scale : sequence
            Length-4 sequence containing the scale factors for each spin channel.
        bkg : sequence
            Length-4 sequence containing the backgrounds for each spin channel.
        """
        R_union = np.zeros(len(x))

        # work out overall SLD of slabs
        sld_u = slabs[:, 1] + slabs[:, 4] * np.sin(np.radians(slabs[:, 5]))
        sld_d = slabs[:, 1] - slabs[:, 4] * np.sin(np.radians(slabs[:, 5]))
        slabs_u = slabs.copy()[:, :4]
        slabs_d = slabs.copy()[:, :4]
        slabs_u[:, 1] = sld_u
        slabs_d[:, 1] = sld_d

        if active_spins[0]:
            # UU
            msk = np.isfinite(x[:, 0])
            xuu = x[msk, 0]
            if isinstance(x_err, np.ndarray):
                xuu_err = x_err[msk, 0]
            else:
                xuu_err = x_err

            _R = reflectivity(
                xuu,
                slabs_u,
                dq=xuu_err,
                threads=self.threads,
                quad_order=self.quad_order,
            )
            R_union[msk] = _R * scale[0] + bkg[0]

        if active_spins[3]:
            # DD
            msk = np.isfinite(x[:, 3])
            xdd = x[msk, 3]
            if isinstance(x_err, np.ndarray):
                xdd_err = x_err[msk, 3]
            else:
                xdd_err = x_err

            _R = reflectivity(
                xdd,
                slabs_d,
                dq=xdd_err,
                threads=self.threads,
                quad_order=self.quad_order,
            )
            R_union[msk] = _R * scale[3] + bkg[3]

        return R_union


def pnr_data_and_generative(objective):
    """
    The Data1D and generative model for a Polarised Reflectometry system

    Parameters
    ----------
    objective : Objective

    Returns
    -------
    pdg : tuple
        Tuple of (Data1D, generative)
    """
    pdg = []
    combined = objective.data
    generative = objective.generative()
    pos = 0
    for spin in combined.spins.keys():
        data = getattr(combined, spin)
        if data is not None:
            npnts = len(data)
            g = generative[pos : pos + npnts]
            pdg.append((data, g))
            pos += npnts
    return pdg



import io
from importlib import resources
from pathlib import Path
import os
from datetime import datetime

from io import BytesIO, StringIO
import pytest

import orsopy.fileio as fio
from orsopy.fileio.base import ValueRange

import refnx
from refnx.dataset import (
    ReflectDataset,
    Data1D,
    load_data,
    OrsoDataset,
    PolarisedReflectDatasets,
)
from refnx.dataset.data1d import _data1D_to_hdf, _hdf_to_data1d
from refnx.dataset.reflectdataset import load_orso
from refnx.reflect import SLD, MaterialSLD, ReflectModel
from refnx.reduce import PlatypusNexus, ReductionOptions
from refnx.reduce.platypusnexus import calculate_wavelength_bins
from refnx.util import q, EPdiv
import refnx.dataset.tests
from refnx._lib import possibly_open_file
import numpy as np
from numpy.testing import assert_equal


class TestReflectDataset:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        self.pth = resources.files(refnx.dataset.tests)

        data = ReflectDataset()

        x1 = np.linspace(0, 10, 115)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data.add_data((x1, y1, e1, dx1))
        self.data = data

        self.cwd = Path.cwd()
        self.tmp_path = tmp_path
        os.chdir(self.tmp_path)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_load(self):
        # load dataset from XML, via file handle
        dataset = ReflectDataset()
        with open(self.pth / "c_PLP0000708.xml") as f:
            dataset.load(f)

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from XML, via string
        dataset = ReflectDataset()
        dataset.load(self.pth / "c_PLP0000708.xml")

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from XML, via string
        dataset = ReflectDataset(self.pth / "c_PLP0000708.xml")

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from .dat, via file handle
        dataset1 = ReflectDataset()
        with open(self.pth / "c_PLP0000708.dat") as f:
            dataset1.load(f)

        assert_equal(len(dataset1), 90)
        assert_equal(90, np.size(dataset1.x))

        # load dataset from .dat, via string
        dataset2 = ReflectDataset()
        dataset2.load(self.pth / "c_PLP0000708.dat")

        assert_equal(len(dataset2), 90)
        assert_equal(90, np.size(dataset2.x))

    def test_dot(self):
        # test with file formats from http://www.reflectometry.net/refdata.htm
        dataset1 = ReflectDataset()
        with open(self.pth / "dot.dat") as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 31)
        assert dataset1.y_err is not None
        assert dataset1.x_err is None

        with open(self.pth / "dot.aft") as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 36)
        assert dataset1.y_err is not None

        with open(self.pth / "dot.refl") as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 59)
        assert dataset1.y_err is not None

        with open(self.pth / "dot.out") as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 1122)
        assert dataset1.y_err is not None
        assert dataset1.x_err is not None

        with open(self.pth / "dot.mft") as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 213)
        assert dataset1.y_err is not None
        assert dataset1.x_err is not None

    def test_load_dat_with_header(self):
        # check that the file load works with a header
        a = ReflectDataset(self.pth / "c_PLP0000708.dat")
        b = ReflectDataset(self.pth / "c_PLP0000708_header.dat")
        c = ReflectDataset(self.pth / "c_PLP0000708_header2.dat")
        assert_equal(len(a), len(b))
        assert_equal(len(a), len(c))

    def test_GH236(self):
        assert (self.pth / "c_PLP0033831.txt").exists()
        a = ReflectDataset(self.pth / "c_PLP0033831.txt")
        assert_equal(len(a), 166)

    def test_loading_junk(self):
        # if you can't load anything from a datafile then you should get a
        # RuntimeError raised.
        from pytest import raises

        with raises(RuntimeError):
            ReflectDataset(self.pth / "../__init__.py")

    def test_bytes(self):
        with open(self.pth / "c_PLP0033831.txt", "rb") as f:
            data = f.read()
        s = data.decode().replace("\r", "\n")
        b = io.StringIO(s)
        d = Data1D()
        d.load(b)

    def test_construction(self):
        # test we can construct a dataset directly from a file.
        pth = self.pth / "c_PLP0000708.xml"

        ReflectDataset(pth)

        with open(self.pth / "c_PLP0000708.xml") as f:
            ReflectDataset(f)

        ReflectDataset(str(self.pth / "d_a.txt"))

    def test_init_with_data1d(self):
        # test we can construct a dataset from a dataset (i.e. a copy)
        dataset = Data1D(self.data)
        assert_equal(dataset.y, self.data.y)

    def test_add_data(self):
        # test we can add data to the dataset

        # 2 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        data.add_data((x1, y1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        data.add_data((x2, y2), requires_splice=True)

        assert len(data) == 13

        # 3 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        data.add_data((x1, y1, e1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        data.add_data((x2, y2, e2), requires_splice=True)

        assert len(data) == 13

        # 4 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data.add_data((x1, y1, e1, dx1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        dx2 = np.ones_like(x2)

        data.add_data((x2, y2, e2, dx2), requires_splice=True)

        assert len(data) == 13

        # test addition of datasets.
        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data = Data1D((x1, y1, e1, dx1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        dx2 = np.ones_like(x2)
        data2 = Data1D((x2, y2, e2, dx2))

        data3 = data + data2
        assert len(data3) == 13

        # test iadd of datasets
        data += data2
        assert len(data) == 13

    def test_add_non_overlapping(self):
        # check that splicing of non-overlapping datasets raises ValueError
        x1 = np.linspace(0, 10, 11)
        y1 = 2 * x1
        data = Data1D((x1, y1))

        x2 = np.linspace(11, 20, 10)
        y2 = 2 * x2 + 2

        from pytest import raises

        with raises(ValueError):
            data.add_data((x2, y2), requires_splice=True)

    def test_save_xml(self):
        self.data.save_xml("test.xml")
        with open("test.xml", "wb") as f:
            self.data.save_xml(f)

    def test_data_setter(self):
        # check that setting a Data1D's data from a tuple works correctly
        # this is the approach used in Data1D.synthesise.
        new_dataset = Data1D()
        new_dataset.data = self.data.data
        assert_equal(new_dataset.y_err, self.data.y_err)
        assert_equal(new_dataset.x_err, self.data.x_err)
        assert_equal(new_dataset.y, self.data.y)
        assert_equal(new_dataset.x, self.data.x)
        assert_equal(new_dataset.weighted, self.data.weighted)

    def test_synthesise(self):
        # add gaussian noise to a Data1D.y
        new_dataset = self.data.synthesise(random_state=1)

        # y-array should not be the same, so don't test that
        assert_equal(new_dataset.y_err, self.data.y_err)
        assert_equal(new_dataset.x_err, self.data.x_err)
        assert_equal(new_dataset.x, self.data.x)
        assert_equal(new_dataset.weighted, self.data.weighted)

        # synthesis should be repeatable with provision of a seed
        new_dataset2 = self.data.synthesise(random_state=1)
        assert_equal(new_dataset.y, new_dataset2.y)

    def test_mask(self):
        # if you mask all points there should be none left
        self.data.mask = np.full_like(self.data.y, False, bool)
        assert_equal(len(self.data), 0)

        # try masking a random selection
        rando = np.random.randint(0, 2, self.data._y.size)
        self.data.mask = rando
        assert_equal(len(self.data), np.count_nonzero(rando))

        # now clear
        self.data.mask = None
        assert_equal(len(self.data), self.data._y.size)

    def test_repr(self):
        a = Data1D(self.pth / "c_PLP0033831.txt")
        b = eval(repr(a))
        assert_equal(len(a), 166)
        assert_equal(len(b), 166)

        # load dataset from XML, via string
        a = ReflectDataset(str(self.pth / "c_PLP0000708.xml"))
        b = eval(repr(a))
        assert_equal(len(b), len(a))
        assert_equal(len(b), len(a))


class TestData1D:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        self.pth = resources.files(refnx.dataset.tests)
        self.cwd = Path.cwd()
        self.tmp_path = tmp_path
        os.chdir(self.tmp_path)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_2D_data(self):
        # despite the class being called Data1D it should be possible to
        # store ND data in it

        # try with multidimensional x data
        x = np.linspace(0.01, 0.5, 100).reshape(50, 2)
        y = np.arange(50, dtype=float)

        data = Data1D((x, y))
        assert len(data) == 50

        assert_equal(data.x, x)
        assert_equal(data.y, y)

        mask = np.ones_like(y, bool)
        mask[1] = False
        data.mask = mask
        assert len(data) == 49
        assert_equal(data.y, np.r_[y[0], y[2:]])
        assert_equal(data.x, np.concatenate((x[0][np.newaxis, :], x[2:])))

        # try with storing multiple channels for y as well
        # this might happen with something like ellipsometry data where you'd
        # have (delta, xsi) as the y data, and (lambda, aoi) as the x data
        # NOTE THAT MASKING WON'T WORK, FOR ND YDATA. This is because the class
        # is using fancy indexing with booleans to remove entries that are
        # unwanted. Could mark as NaN instead?
        x = np.linspace(0.01, 0.5, 50).reshape(50)
        y = np.arange(100, dtype=float).reshape(50, 2)

        data = Data1D((x, y))
        assert len(data) == 100

        assert_equal(data.x, x)
        assert_equal(data.y, y)

        x = np.linspace(0.01, 0.5, 100).reshape(50, 2)

        data = Data1D((x, y))
        assert len(data) == 100

        assert_equal(data.x, x)
        assert_equal(data.y, y)

    def test_rudimentry_HDF(self):
        rng = np.random.default_rng()
        x = np.linspace(0, 100, 101)
        y = rng.random(101)
        dy = rng.random(101)
        dx = rng.random((101, 2, 40))

        data = Data1D((x, y, dy, dx))

        f = BytesIO()
        _data1D_to_hdf(f, data)
        _data = _hdf_to_data1d(f)
        assert len(_data) == 101
        assert _data.x_err.shape == (101, 2, 40)

    def test_load_data(self):
        # test the load_data function by trying to load all the files in the
        # test directory
        fs = Path.cwd().glob("*.*")
        fs = [f for f in fs if not f.endswith(".py")]
        fs = [f for f in fs if not f.startswith("coef_")]

        for f in fs:
            load_data(f)
            load_data(Path(f))


class TestOrtDataset:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path, data_directory):
        self.pth = resources.files(refnx.dataset.tests)
        self.cwd = Path.cwd()
        self.tmp_path = tmp_path
        self.data_directory = data_directory
        os.chdir(self.tmp_path)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_OrtDataset_init(self):
        f = self.pth / "ORSO_data.ort"
        # create with path
        OrsoDataset(f)

        # create with orsopy.fileio.orso.OrsoDataset object
        OrsoDataset(load_orso(f)[0])

        # create with orsopy.fileio.orso.OrsoDataset object
        ds = OrsoDataset(
            load_orso(self.data_directory / "dataset" / "Ni_example.ort")[0]
        )
        assert ds.orso is not None
        assert len(ds) == 341
        # checks sd --> fwhm conversion of resolution function
        np.testing.assert_allclose(ds.x_err, 0.05 * ds.x, rtol=0.005)

    def test_setup_analysis(self):
        # create with orsopy.fileio.orso.OrsoDataset object
        ds = OrsoDataset(
            load_orso(self.data_directory / "dataset" / "Ni_example.ort")[0]
        )
        s, model, objective = ds.setup_analysis()
        np.testing.assert_allclose(s[1].thick.value, 1000)

        # now try changing value and resaving ort file
        s[1].thick.value = 2000.0
        ds.update_model(s)
        ds.save(self.tmp_path / "flake.ort")

        # see if the value was updated in the file
        ds2 = OrsoDataset(load_orso(self.tmp_path / "flake.ort")[0])
        s2, model2, objective2 = ds2.setup_analysis()
        np.testing.assert_allclose(s2[1].thick.value, 2000)

    def test_ort_load(self):
        f = self.pth / "ORSO_data.ort"
        assert f.exists()
        try:
            load_orso(str(f))
            load_orso(f)
        except ImportError:
            # load_orso had problems on Python 3.10, so bypass the test
            return

        d = load_data(f)
        assert len(d) == 2
        assert isinstance(d, OrsoDataset)
        d.refresh()

        # no model is associated with this ORT file, so one cannot create
        # a ReflectModel from it.
        with pytest.raises(RuntimeError):
            d.setup_analysis()

    def test_orb_load(self):
        f = self.pth / "test.orb"
        assert f.exists()

        try:
            load_orso(str(f))
            load_orso(f)
        except ImportError:
            # load_orso had problems on Python 3.10, so bypass the test
            return

        # with possibly_open_file(f, 'rb') as f:
        d = load_data(f)

        assert isinstance(d, OrsoDataset)
        d.refresh()

    def test_create_example_ort_from_simulation(self, data_directory):
        air = SLD(0)
        sio2 = SLD(3.47)
        ni = MaterialSLD("Ni", density=8.9)
        si = SLD(2.07)

        dq = 5.0

        s = air | ni(1000, 4) | sio2(10, 3) | si(0, 3.5)
        model = ReflectModel(s, bkg=5e-7, dq=dq)

        # Code for rough simulation of data based on Hogben by Jos Cooper
        # Resolution smearing from the instrument is 'perfectly gaussian',
        # and is carried out by the ReflectModel.
        # A more sophisticated (but slower) approach to resolution smearing
        # is available.
        rng = np.random.default_rng(121908290)

        # 0.65 specifies at what configuration the direct beam was measured at.
        with pytest.warns(RuntimeWarning, match="specular ridge search"):
            h = Hoggy(
                model,
                0.65,
                self.data_directory / "reduce" / "PLP0049278.nx.hdf",
                attenuator=28,
            )

        # two angles of incidence simulated, 0.8 and 3.5
        ds = h(0.8, 120, rng=rng)
        ds1 = h(3.5, 300, rng=rng)

        ds += ds1

        header = fio.orso.Orso.empty()
        header.data_source.owner = fio.data_source.Person(
            name="Joe Bloggs", affiliation="Unseen University"
        )
        header.data_source.experiment = fio.data_source.Experiment(
            title="Metal films",
            start_date=datetime(2025, 4, 8),
            instrument="Platypus",
            probe="neutron",
            facility="ANSTO",
            proposalID="1234",
        )
        header.data_source.sample = fio.data_source.Sample(
            name="Ni on Si",
            category="from air",
            description="~1000 A of metal",
        )
        header.data_source.measurement = fio.data_source.Measurement(
            fio.data_source.InstrumentSettings(
                incident_angle=ValueRange(
                    min=0.8, max=3.5, individual_magnitudes=[0.8, 3.5]
                ),
                wavelength=ValueRange(min=2.8, max=19),
            ),
            data_files=[
                "PLP000001.nx.hdf",
                "PLP000002.nx.hdf",
                "PLP0049278.nx.hdf",
                "PLP0049278.nx.hdf",
            ],  # spoofed
        )

        q_column = fio.base.Column(
            name="Qz",
            unit="1/angstrom",
            physical_quantity="wavevector transfer",
        )
        r_column = fio.base.Column(
            name="R", unit=None, physical_quantity="reflectivity"
        )
        dr_column = fio.base.ErrorColumn(
            error_of="R", error_type="uncertainty", value_is="sigma"
        )
        dq_column = fio.base.ErrorColumn(
            error_of="Qz", error_type="resolution", value_is="sigma"
        )

        header.columns = [q_column, r_column, dr_column, dq_column]

        dataset = fio.orso.OrsoDataset(info=header, data=np.array(ds.data).T)
        dataset.save(self.tmp_path / "Ni_example.ort")


class Hoggy:
    def __init__(
        self,
        model,
        angle_scale,
        spectrum,
        attenuator=1.0,
        rebin_percent=1.0,
        lo_wavelength=2.8,
        hi_wavelength=19.0,
    ):
        self.model = model
        self.w_bins = calculate_wavelength_bins(
            lo_wavelength, hi_wavelength, rebin_percent
        )
        self.wval = 0.5 * (self.w_bins[1:] + self.w_bins[:-1])
        self.angle_scale = angle_scale
        self.attenuator = attenuator

        rdo = ReductionOptions(
            lo_wavelength=lo_wavelength,
            hi_wavelength=hi_wavelength,
            background=False,
            normalise=False,
            normalise_bins=False,
            rebin_percent=rebin_percent,
        )

        pn = PlatypusNexus(spectrum)
        self.measure_time_direct = pn.cat.cat["time"]

        # direct beam spectrum
        lam, i, di = pn.process(**rdo)
        self.lam = np.squeeze(lam)
        self.i = np.squeeze(i)
        self.di = np.squeeze(di)
        self.spec = np.copy(self.i)

        # divide by count time to get cps on direct beam
        self.spec /= self.measure_time_direct[0]
        self.spec *= attenuator

    def __call__(self, angle, measure_time, rng=None):
        if rng is None:
            rng = np.random.default_rng()

        q_bins = q(angle, self.w_bins)
        qval = 0.5 * (q_bins[1:] + q_bins[:-1])

        # scale angle compared to incident beam, assuming slits are opened proportionally
        scalar = np.pow(angle / self.angle_scale, 2)
        _spec = self.spec * scalar

        # scaled spectrum for actual reflectivity measurement
        incident_during_reflect = _spec * measure_time
        reflected_counts = rng.poisson(
            incident_during_reflect * self.model(qval)
        )
        dreflected_counts = np.sqrt(reflected_counts)

        reflectivity, dreflectivity = EPdiv(
            reflected_counts, dreflected_counts, self.i, self.di
        )
        reflectivity *= (self.measure_time_direct) / (
            self.attenuator * measure_time * scalar
        )
        dreflectivity *= (self.measure_time_direct) / (
            self.attenuator * measure_time * scalar
        )

        dataset = Data1D(
            data=(
                qval,
                reflectivity,
                dreflectivity,
                self.model.dq.value / 100 / 2.3548 * qval,
            )
        )
        return dataset


class TestPolarisedReflectDatasets:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        self.pth = resources.files(refnx.reflect.tests)
        self.cwd = Path.cwd()
        self.tmp_path = tmp_path
        os.chdir(self.tmp_path)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_properties(self):
        data_uu = Data1D(self.pth / "c_PLP0007885.dat")
        data_dd = Data1D(self.pth / "c_PLP0007885.dat")
        combined = PolarisedReflectDatasets(down_down=data_dd, up_up=data_uu)
        y = combined.y
        x = combined.x
        y_err = combined.y_err
        x_err = combined.x_err

        assert_equal(x.shape, x_err.shape)
        assert_equal(y.shape, y_err.shape)
        assert combined.weighted
        assert len(y) == len(x)



import string
import time
import re

# from datetime import datetime
from pathlib import Path, PurePath

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
import numpy as np

import orsopy.fileio as fio
from refnx.dataset import Data1D
from refnx._lib import possibly_open_file


def load_orso(f):
    if isinstance(f, fio.orso.OrsoDataset):
        return [f]
    try:
        return fio.load_orso(f)
    except Exception:
        pass
    try:
        v = fio.load_nexus(f)
        return v
    except Exception:
        pass


_template_ref_xml = """<?xml version="1.0"?>
<REFroot xmlns="">
<REFentry time="$time">
<Title>$title</Title>
<User>$user</User>
<REFsample>
<ID>$sample</ID>
</REFsample>
<REFdata axes="Qz" rank="1" type="POINT" spin="UNPOLARISED" dim="$numpoints">
<Run filename="$datafilenumber" preset="" size="">
</Run>
<R uncertainty="dR">$_ydata</R>
<Qz uncertainty="dQz" units="1/A">$_xdata</Qz>
<dR type="SD">$_ydataSD</dR>
<dQz type="_FWHM" units="1/A">$_xdataSD</dQz>
</REFdata>
</REFentry>
</REFroot>"""


class ReflectDataset(Data1D):
    """
    A 1D Reflectivity dataset.
    """

    def __init__(self, data=None, **kwds):
        """
        Initialise a reflectivity dataset.

        Parameters
        ----------
        data : {str, file-like, Path, tuple of np.ndarray} optional
            `data` can be a string, file-like, or Path object referring to a File to
            load the dataset from.

            Alternatively it is a tuple containing the data from which the
            dataset will be constructed. The tuple should have between 2 and 4
            members.

                - data[0] - Q
                - data[1] - R
                - data[2] - dR
                - data[3] - dQ

            `data` must be at least two long, `Q` and `R`.
            If the tuple is at least 3 long then the third member is `dR`.
            If the tuple is 4 long then the fourth member is `dQ`.
            All arrays must have the same shape.
        """
        super().__init__(data=data, **kwds)
        self.datafilenumber = list()
        self.sld_profile = None

    def __repr__(self):
        msk = self._mask
        if np.all(self._mask):
            msk = None

        if self.filename is not None:
            return f"ReflectDataset(data={str(self.filename)!r}, mask={msk!r})"
        else:
            return f"ReflectDataset(data={self.data!r}, mask={msk!r})"

    def save_xml(self, f, start_time=0):
        """
        Saves the reflectivity data to an XML file.

        Parameters
        ----------
        f : str or file-like
            The file to write the spectrum to, or a str that specifies the file
            name
        start_time: int, optional
            Epoch time specifying when the sample started
        """
        s = string.Template(_template_ref_xml)
        self.time = time.strftime(
            "%Y-%m-%dT%H:%M:%S", time.localtime(start_time)
        )
        # self.time = time.strftime(
        # datetime.fromtimestamp(start_time).isoformat()
        # filename = 'c_PLP{:07d}_{:d}.xml'.format(self._rnumber[0], 0)

        self._ydata = repr(self.y.tolist()).strip(",[]")
        self._xdata = repr(self.x.tolist()).strip(",[]")
        self._ydataSD = repr(self.y_err.tolist()).strip(",[]")
        self._xdataSD = repr(self.x_err.tolist()).strip(",[]")

        thefile = s.safe_substitute(self.__dict__)

        with possibly_open_file(f, "wb") as g:
            if "b" in g.mode:
                thefile = thefile.encode("utf-8")

            g.write(thefile)

    def load(self, f):
        """
        Load a dataset from file. Can either be 2-4 column ascii or XML file.

        Parameters
        ----------
        f : {str, file-like, Path}
            The file to load the spectrum from, or a str that specifies the
            file name
        """
        if hasattr(f, "read") and hasattr(f, "write"):
            if hasattr(f, "name"):
                # file-like ?
                fname = f.name
            else:
                fname = ""
        else:
            fname = f
        try:
            tree = ET.ElementTree()
            tree.parse(f)

            delim = ", | |,"
            qtext = re.split(delim, tree.find(".//Qz").text)
            rtext = re.split(delim, tree.find(".//R").text)
            drtext = re.split(delim, tree.find(".//dR").text)
            dqtext = re.split(delim, tree.find(".//dQz").text)

            qvals = [float(val) for val in qtext if len(val)]
            rvals = [float(val) for val in rtext if len(val)]
            drvals = [float(val) for val in drtext if len(val)]
            dqvals = [float(val) for val in dqtext if len(val)]

            if isinstance(fname, PurePath):
                # use a PurePath, not a system specific path type
                # because Posix systems can't deal with WindowsPath
                # and vice versa. This becomes an issue when pickling.
                fname = PurePath(fname)

            self.filename = fname
            self.name = Path(fname).stem
            self.data = (qvals, rvals, drvals, dqvals)
        except ET.ParseError:
            super().load(fname)


class OrsoDataset(Data1D):
    """
    A thinly wrapped version of an ORSODataset

    Parameters
    ----------
    data : {str, file-like, Path}

    Notes
    -----
    Multiplies the resolution information contained in the fourth column
    of the ORSO dataset to convert from standard deviation to FWHM.
    """

    def __init__(self, data=None, **kwds):
        super().__init__(data=None, **kwds)
        self._orso = None
        if data is not None:
            if isinstance(data, fio.orso.OrsoDataset):
                self.orso = [data]
            else:
                self.load(data)

    @property
    def orso(self):
        return self._orso

    @orso.setter
    def orso(self, value):
        if isinstance(value[0], fio.orso.OrsoDataset):
            self._orso = [value[0]]
            self._set_internals()

    def _set_internals(self):
        """
        Updates internal information after an OrsoDataset has been associated
        with the object
        """
        header = self.orso[0].info

        _data = self.orso[0].data[:, :4].T

        # figure out if data was in 1/nm or 1/angstrom
        # internally refnx uses reciprocal angstrom
        q_units = header.columns[0].unit.lower()
        if q_units == "1/nm":
            # need to divide q by 10
            _data[0] /= 10.0

            if _data.shape[0] > 3:
                _data[3] /= 10.0

        # ORSO files save resolution information as SD,
        # internally refnx uses FWHM
        if _data.shape[0] > 3:
            _data[3] *= 2.3548

        self.data = _data

    def load(self, f):
        """
        Parameters
        ----------
        f : {str, file-like, Path}
            The file to load the spectrum from, or a str/Path that specifies
            the file name
        """
        if hasattr(f, "read") and hasattr(f, "write"):
            if hasattr(f, "name"):
                # file-like ?
                fname = f.name
            else:
                fname = ""
        else:
            # string-like??
            fname = str(f)

        mode = "r"
        if fname.endswith(".orb"):
            mode = "rb"

        with possibly_open_file(f, mode) as g:
            self.orso = load_orso(g)

        self.filename = fname
        self.name = Path(fname).stem

    def save(self, f):
        """
        Saves the dataset to an ORT file.

        Parameters
        ----------
        f : {file-like, str, Path}
            File to save the dataset to.

        """
        self.orso[0].save(f)

    def refresh(self):
        """
        Refreshes a previously loaded dataset.

        """
        # OrsoDataset needs to carry its own implementation
        # opening a binary file needs to be done with correct mode
        # vs opening a text file.
        if self.filename is not None:
            self.load(self.filename)

    def setup_analysis(self):
        """
        Creates a Structure, ReflectModel, Objective from the information
        contained within the ORSO file.

        Returns
        -------
        (s, model, objective) : tuple
            tuple comprising the Structure, ReflectModel, and Objective
            created from the Orso file.

        Notes
        -----
        If no model structure is specified in the OrsoDataset then a
        RuntimeError will be raised.

        Example
        -------

        >>> import urllib.request
        >>> import shutil
        >>> url = "https://github.com/refnx/refnx-testdata/raw/refs/heads/master/data/dataset/Ni_example.ort"
        >>> with (urllib.request.urlopen(url, timeout=5) as response, open("Ni_example.ort", 'wb') as f):
        ...     shutil.copyfileobj(response, f)
        >>> from refnx.dataset import OrsoDataset
        >>> ds = OrsoDataset("Ni_example.ort")
        >>> s, model, objective = ds.setup_analysis()
        >>> print(model)
        >>> s[1].thick.setp(vary=True, bounds=(990, 1010))
        >>> from refnx.analysis import CurveFitter
        >>> fitter = CurveFitter(objective)
        >>> fitter.fit("differential_evolution")

        """
        from refnx.reflect import Structure, ReflectModel
        from refnx.analysis import Objective

        if self.orso is None:
            raise RuntimeError(
                "This instance has not yet been initialised"
                " with an OrsoDataset"
            )

        model = self.orso[0].info.data_source.sample.model
        if model is None:
            raise RuntimeError("No model is associated with the OrsoDataset")

        s = Structure.from_orso(model)
        model = ReflectModel(s)
        objective = Objective(model, self)
        return s, model, objective

    def update_model(self, structure):
        """
        Updates the model in the OrsoDataset from a given Structure.

        Parameters
        ----------
        structure : refnx.reflect.Structure

        Notes
        -----
        Only simple Slab structures can be processed at this time.
        """
        if self.orso is None:
            raise RuntimeError(
                "This instance has not yet been initialised"
                " with an OrsoDataset"
            )

        # figure out ORSO model from the Structure
        model = structure.to_orso()

        # now insert into the OrsoDataset
        self.orso[0].info.data_source.sample.model = model


class PolarisedReflectDatasets:
    def __init__(self, down_down=None, up_up=None, down_up=None, up_down=None):
        names = []
        for o in (down_down, up_up, down_up, up_down):
            if o is not None:
                if not isinstance(o, Data1D):
                    raise TypeError(f"{o} is not a Data1D object.")
                names.append(o.name)
            self.name = ", ".join(names)

        self.down_down = down_down
        self.up_up = up_up
        self.down_up = down_up
        self.up_down = up_down
        self.spins = {"up_up": 0, "up_down": 1, "down_up": 2, "down_down": 3}

    @property
    def x(self):
        xs = []
        for spin in self.spins.keys():
            data = getattr(self, spin)
            if data is None:
                continue
            else:
                full = np.full((len(data.x), 4), np.nan)
                full[:, self.spins[spin]] = data.x
                xs.append(full)

        return np.r_[xs].reshape(-1, 4)

    @property
    def y(self):
        ys = []
        for spin in self.spins.keys():
            data = getattr(self, spin)
            if data is None:
                continue
            else:
                ys.append(data.y)

        return np.concatenate(ys)

    @property
    def y_err(self):
        if self.weighted:
            ys = []
            for spin in self.spins.keys():
                data = getattr(self, spin)
                if data is None:
                    continue
                else:
                    ys.append(data.y_err)

            return np.concatenate(ys)
        else:
            return None

    @property
    def x_err(self):
        xs = []
        for spin in self.spins.keys():
            data = getattr(self, spin)
            if data is None:
                continue
            else:
                if data.x_err is None:
                    return None
                full = np.full((len(data.x), 4), np.nan)
                full[:, self.spins[spin]] = data.x_err
                xs.append(full)

        return np.r_[xs].reshape(-1, 4)

    @property
    def data(self):
        return (self.x, self.y, self.y_err, self.x_err)

    @property
    def weighted(self):
        weighted = []
        for spin in self.spins.keys():
            data = getattr(self, spin)
            if data is not None:
                weighted.append(data.weighted)
        return all(weighted)

    def __len__(self):
        return len(self.y)



# -*- coding: utf-8 -*-
from functools import lru_cache

import numpy as np

from .red_blue import RedBlueMove

__all__ = ["DEMove"]


class DEMove(RedBlueMove):
    r"""A proposal using differential evolution.

    This `Differential evolution proposal
    <http://www.stat.columbia.edu/~gelman/stuff_for_blog/cajo.pdf>`_ is
    implemented following `Nelson et al. (2013)
    <https://doi.org/10.1088/0067-0049/210/1/11>`_.

    Args:
        sigma (float): The standard deviation of the Gaussian used to stretch
            the proposal vector.
        gamma0 (Optional[float]): The mean stretch factor for the proposal
            vector. By default, it is :math:`2.38 / \sqrt{2\,\mathrm{ndim}}`
            as recommended by the two references.

    """

    def __init__(self, sigma=1.0e-5, gamma0=None, **kwargs):
        self.sigma = sigma
        self.gamma0 = gamma0
        super().__init__(**kwargs)

    def setup(self, coords):
        self.g0 = self.gamma0
        if self.g0 is None:
            # Pure MAGIC:
            ndim = coords.shape[1]
            self.g0 = 2.38 / np.sqrt(2 * ndim)

    def get_proposal(self, s, c, random):
        c = np.concatenate(c, axis=0)
        ns, ndim = s.shape
        nc = c.shape[0]

        # Get the pair indices
        pairs = _get_nondiagonal_pairs(nc)

        # Sample from the pairs
        indices = random.choice(pairs.shape[0], size=ns, replace=True)
        pairs = pairs[indices]

        # Compute diff vectors
        diffs = np.diff(c[pairs], axis=1).squeeze(axis=1)  # (ns, ndim)

        # Sample a gamma value for each walker following Nelson et al. (2013)
        gamma = self.g0 * (1 + self.sigma * random.randn(ns, 1))  # (ns, 1)

        # In this way, sigma is the standard deviation of the distribution of gamma,
        # instead of the standard deviation of the distribution of the proposal as proposed by Ter Braak (2006).
        # Otherwise, sigma should be tuned for each dimension, which confronts the idea of affine-invariance.

        q = s + gamma * diffs

        return q, np.zeros(ns, dtype=np.float64)


@lru_cache(maxsize=1)
def _get_nondiagonal_pairs(n: int) -> np.ndarray:
    """Get the indices of a square matrix with size n, excluding the diagonal."""
    rows, cols = np.tril_indices(n, -1)  # -1 to exclude diagonal

    # Combine rows-cols and cols-rows pairs
    pairs = np.column_stack(
        [np.concatenate([rows, cols]), np.concatenate([cols, rows])]
    )

    return pairs



"""
Auto reduce reflectometry files
"""
#!/usr/bin/env python
import os
import sys
import time
import argparse
from refnx.reduce import AutoReducer


def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Auto-reduce reflectometry files.")
    parser.add_argument("file_list", nargs='+')
    parser.add_argument("-p", "--path", type=dir_path, default="./", help="path to datafiles")
    parser.add_argument("-s", "--scale", help="scale factor",
                        type=float, default=1.0)
    args = parser.parse_args()
    files = args.file_list
    pth = args.path
    files = [os.path.join(pth, file) for file in files]
    print(f"Path: {pth}")
    print(f"Reducing against: {files}, with scale: {args.scale}")
    ar = AutoReducer(files, data_folder=pth, scale=args.scale)

    while True:
        time.sleep(10.)



#!/usr/bin/env python
"""
global_fitting_from_motofit.py [OPTIONS] [-- ARGS]

Sets and executes up a global fitting environment supplied by Motofit in IGOR.

Examples::

    $ python global_fitting_from_motofit.py global_pilot_file

"""
import sys
import numbers
import time
from argparse import ArgumentParser
from copy import deepcopy
from multiprocessing import Pool

import numpy as np
from refnx.analysis import (CurveFitter, ReflectivityFitFunction, GlobalFitter,
                            to_parameters, Transform, values, names)
from refnx.dataset import ReflectDataset


def global_fitter_setup(global_pilot_file, dqvals=5.0):
    # Parse the global_fitter setup from Igor.
    # TODO deal with user generated non-slab models.
    with open(global_pilot_file, 'r') as f:
        data_files = f.readline().split()
        pilot_files = f.readline().split()

    constraints = np.loadtxt(global_pilot_file, skiprows=2, dtype=int)

    # open the datafiles
    datasets = []
    for data_file in data_files:
        dataset = ReflectDataset(data_file)
        datasets.append(dataset)

    # deal with the individual pilot files
    parameters = []
    for pilot_file in pilot_files:
        pars = np.loadtxt(pilot_file, skiprows=4)

        # lets just assume for now that the data has resolution info
        # and that we're doing a slab model.
        pv = pars[:, 0][:]
        varies = (pars[:, 1].astype(int) == 0)

        # workout bounds, and account for the fact that MotofitMPI
        # doesn't set bounds for parameters that are fixed
        bounds = []
        for idx in range(np.size(pv)):
            if not varies[idx]:
                bounds.append((0, 2 * pv[idx]))
            else:
                bounds.append(pars[idx, 2:4])

        P = to_parameters(pv,
                          varies=varies,
                          bounds=bounds)
        parameters.append(P)

    # now create CurveFitting instances
    T = Transform('logY')
    fitters = []

    for parameter, dataset in zip(parameters, datasets):
        t_data_y, t_data_yerr = T.transform(dataset.x,
                                            dataset.y,
                                            dataset.y_err)

        if isinstance(dqvals, numbers.Real):
            _dqvals = float(dqvals)
        else:
            _dqvals = dataset.x_err

        c = CurveFitter(ReflectivityFitFunction(T.transform, workers=True),
                        (dataset.x, t_data_y, t_data_yerr),
                        parameter,
                        fcn_kws={'dqvals': _dqvals})
        fitters.append(c)

    # create globalfitter
    # setup constraints
    unique, indices = np.unique(constraints, return_index=True)

    # TODO assertions for checking linkage integrity

    n_datasets = len(datasets)

    def is_unique(row, col):
        ravelled_idx = row * n_datasets + col
        return ravelled_idx in indices

    cons = []
    for col in range(n_datasets):
        for row, val in enumerate(parameters[col]):
            if constraints[row, col] == -1 or is_unique(row, col):
                continue
            # so it's not unique, but which parameter does it depend on?
            # find location of master parameter
            master = np.extract(unique == constraints[row, col], indices)[0]
            m_col = master % n_datasets
            m_row = (master - m_col) // n_datasets
            constraint = 'd%u:p%u = d%u:p%u' % (col, row, m_col, m_row)
            cons.append(constraint)

            # also have to rejig the bounds because MotoMPI doesn't
            # set bounds for those that aren't unique. But this is bad for
            # lmfit because it'll clip them.
            par = fitters[col].params['p%u' % row]
            m_par = fitters[m_col].params['p%u' % m_row]
            par.min = m_par.min
            par.max = m_par.max

    global_fitter = GlobalFitter(fitters, constraints=cons)

    # # update the constraints
    # global_fitter.params.update_constraints()

    return global_fitter


def _mcmc(args, global_fitter):
    # sample via Markov Chain Monte Carlo
    pos = None
    if args.chain_input is not None:
        pos = np.load(args.chain_input)

    # do the sampling
    chunk_size = 50
    n_remaining = args.steps
    done = 0

    sys.stdout.write("----------------------\n")
    sys.stdout.write("Starting MCMC\n")
    sys.stdout.write("----------------------\n")
    start = time.time()

    reuse_sampler = False
    while n_remaining > 0:
        todo = min(chunk_size, n_remaining)

        res = global_fitter.emcee(nwalkers=args.walkers,
                                  steps=todo,
                                  ntemps=args.ntemps,
                                  burn=0,
                                  thin=1,
                                  workers=args.nprocesses,
                                  pos=pos,
                                  reuse_sampler=reuse_sampler)
        reuse_sampler = True
        n_remaining -= todo
        done += todo
        pos = res.chain

        # write raw chain in npy format. It is unburnt and unthinned
        np.save(args.chain_output, res.chain)

        sys.stdout.write(
            "{0:^7} steps, {1:^7} seconds\n".format(done, time.time() - start)
                         )

    # thin and burn the chain.
    chain = res.chain[..., args.burn::args.thin, :]
    res.chain = np.copy(chain)

    return res


def __resample_mc_iterator(args):
    global_fitter, seed = args
    gf = deepcopy(global_fitter)
    np.random.seed(seed)
    res = gf._resample_mc(1, 'differential_evolution')
    return res.mc


def _resample_mc(args, global_fitter):
    # do the sampling by Resampling Monte Carlo

    sys.stdout.write("----------------------\n")
    sys.stdout.write("Starting resampling MC\n")
    sys.stdout.write("----------------------\n")
    start = time.time()

    # do a single fit first
    output = global_fitter.fit('differential_evolution')

    chunksize = 5 * args.nprocesses
    remaining = args.steps
    done = 0
    mcs = []

    with Pool(args.nprocesses) as pool:
        while remaining > 0:
            todo = min(remaining, chunksize)

            # seeding the random number generator
            seeds = range(done, done + todo)
            gfs = [global_fitter] * todo

            results = pool.map(__resample_mc_iterator, zip(gfs, seeds))

            mcs.append(results)
            done += todo
            remaining -= todo

            sys.stdout.write("{0:^7} steps, {1:^7} seconds\n".format(done, time.time() - start))

    mc = np.squeeze(np.vstack(mcs))

    quantiles = np.percentile(mc, [15.87, 50, 84.13], axis=0)

    params = output.params
    for i, var_name in enumerate(output.var_names):
        std_l, median, std_u = quantiles[:, i]
        params[var_name].value = median
        params[var_name].stderr = 0.5 * (std_u - std_l)
        params[var_name].correl = {}

    params.update_constraints()

    # work out correlation coefficients
    corrcoefs = np.corrcoef(mc.T)

    for i, var_name in enumerate(output.var_names):
        for j, var_name2 in enumerate(output.var_names):
            if i != j:
                output.params[var_name].correl[var_name2] = corrcoefs[i, j]

    output.mc = mc
    output.chain = mc
    output.errorbars = True
    output.nvarys = len(output.var_names)

    return output


def main(argv):
    parser = ArgumentParser(usage=__doc__.lstrip())
    parser.add_argument("global_pilot_file",
                        help="The name of the global pilot file")
    parser.add_argument("--walkers", "-w", type=int, default=100,
                        help="How many MCMC walkers? Default=100")
    parser.add_argument("--steps", "-s", type=int, default=2000,
                        help="How many MCMC steps? Default=2000")
    parser.add_argument("--ntemps", '-T', type=int, default=1,
                        help="How many parallel tempering temperatures? "
                             "Default=1")
    parser.add_argument("--burn", "-b", type=int, default=500,
                        help="How many initial MCMC steps do you want to "
                             "burn? Default=500")
    parser.add_argument("--thin", "-t", type=int, default=20,
                        help="Thins the chain by accepting 1 in every 'thin'. "
                             "Default=20")
    parser.add_argument("--resample", "-r", default=False, action='store_true',
                        help="Don't do Markov Chain Monte Carlo, do resampling "
                             "MC instead. All MCMC parameters are ignored if "
                             "this option is specified.")
    parser.add_argument("--qres", "-q", type=float, default=5.0,
                        help="Constant dq/q resolution. Default=5")
    parser.add_argument("--pointqres", "-p", action="store_true",
                        default=False,
                        help="Use point by point resolution smearing. "
                             "Default=False")
    parser.add_argument("--chain_input", "-i", type=str,
                        help="Initialise/restart emcee with this RAW chain. "
                             "This file is a numpy array (.npy) that would've "
                             "originally been saved by the --chain_output "
                             "option.")
    parser.add_argument("--chain_output", "-c", default='raw_chain.npy',
                        type=str,
                        help="Specify filename for unthinned, unburnt RAW "
                             "chain. The file is saved as a numpy (.npy) "
                             "array. You can use this file if you'd like to do "
                             "the burn/thin procedure yourself. You can also"
                             " use this file to restart the sampling. The array"
                             " has shape (walkers, steps, dims), where dims "
                             "represents the number of parameters you are "
                             "varying. If ntemps is > 1 then the array has "
                             "shape (ntemps, walkers, steps, dims). "
                             "Default=raw_chain.npy")
    parser.add_argument("--output", "-o", type=str, default='iterations',
                        help="Output file for burnt and thinned MCMC chain. "
                             "This is only written once the sampling has "
                             "finished")
    parser.add_argument("--nprocesses", "-n", type=int, default=1,
                        help="How many processes for parallelisation? Default=1")

    args = parser.parse_args(argv)

    # set up global fitting
    if args.pointqres:
        dqvals = None
    else:
        dqvals = args.qres

    if args.nprocesses < 1:
        args.nprocesses = 1

    global_fitter = global_fitter_setup(args.global_pilot_file,
                                        dqvals=dqvals)

    # do the Monte Carlo
    if not args.resample:
        # By Markov Chain Monte Carlo
        if args.thin < 1:
            sys.stdout.write("Can't have thin < 1, setting 'thin' to 1.\n")
            args.thin = 1
        if (args.burn < 0) or (args.burn > args.steps):
            sys.stdout.write("Can't burn < 0 or burn > steps, setting 'burn' to 1.\n")
            args.burn = 1

        res = _mcmc(args, global_fitter)
    else:
        # By resampling
        res = _resample_mc(args, global_fitter)

    # write the iterations.
    _write_results(args.output, res)

    sys.stdout.write("\nFinished MCMC\n")
    sys.stdout.write("-------------\n")
    sys.stdout.write(fit_report(res.params))
    sys.stdout.write("\n-----------------------------------------------------\n")


def filter_dependent_params(params, output):
    # filters dependent parameters from the MonteCarlo output
    # i.e. reject all parameters where expr is not None.
    independent = []
    for idx, param in enumerate(params):
        if param.expr is not None:
            independent.append(idx)

    arr = np.zeros((np.size(output, 0), idx))
    arr = output[:, idx]
    return arr


def _write_results(f, emcee_result):
    # the flatchain is what we're interested in.
    # make an output array
    # hopefully the chain has been burned and thinned enough.
    output = np.zeros((np.size(emcee_result.flatchain, 0),
                       len(emcee_result.params)))
    gen = pgen(emcee_result.params, emcee_result.flatchain)
    for row in output:
        pars = next(gen)
        row[:] = values(pars)[:]

    np.savetxt(f, output, header=' '.join(names(emcee_result.params)))


def pgen(parameters, flatchain, idx=None):
    # generator for all the different parameters from a flatchain.
    if idx is None:
        idx = range(np.size(flatchain, 0))
    for i in idx:
        for var_name in flatchain.columns:
            parameters[var_name].value = flatchain.iloc[i][var_name]
        yield parameters


if __name__ == "__main__":
    main(argv=sys.argv[1:])



# cython: language_level=3


def c_flatten(seq):
    """
    Flatten a nested sequence. Ordering is preserved

    Parameters
    ----------
    seq : sequence
        The sequence to flatten

    Returns
    -------
    el : generator
        yields flattened sequences from seq
    """
    for el in seq:
        try:
            iter(el)
            if isinstance(el, (str, bytes)):
                raise TypeError
            yield from c_flatten(el)
        except TypeError:
            yield el


def c_unique(seq, idfun=id):
    """
    List of unique values in sequence (by object id).

    Parameters
    ----------
    seq : sequence
    idfun : callable

    Returns
    -------
    p : generator
        yields unique values from l
    """
    seen = {}
    for item in seq:
        marker = idfun(item)
        if marker not in seen:
            seen[marker] = 1
            yield item



# cython: language_level=3, boundscheck=False
"""
    *Calculates the specular (Neutron or X-ray) reflectivity from a stratified
    series of layers.

The refnx code is distributed under the following license:

Copyright (c) 2015 A. R. J. Nelson, Australian Nuclear Science and Technology Organisation

Permission to use and redistribute the source code or binary forms of this
software and its documentation, with or without modification is hereby
granted provided that the above notice of copyright, these terms of use,
and the disclaimer of warranty below appear in the source code and
documentation, and that none of the names of above institutions or
authors appear in advertising or endorsement of works derived from this
software without specific prior written permission from all parties.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THIS SOFTWARE.

"""
from multiprocessing import cpu_count
from cpython.mem cimport PyMem_Malloc, PyMem_Free
import numpy as np
cimport numpy as np
cimport cython


np.import_array()

cdef extern from "refcaller.h" nogil:
    void abeles_wrapper(
        int numcoefs,
        const double *coefP,
        int npoints,
        double *yP,
        const double *xP
    )
    void abeles_wrapper_MT(
        unsigned int batch,
        int numcoefs,
        const double *coefP,
        int npoints,
        double *yP,
        const double *xP,
        int threads
    )
    void parratt_wrapper(
        int numcoefs,
        const double *coefP,
        int npoints,
        double *yP,
        const double *xP
    )
    void parratt_wrapper_MT(
        unsigned int batch,
        int numcoefs,
        const double *coefP,
        int npoints,
        double *yP,
        const double *xP,
        int threads
    )
    void pnr(
        int layers,
        const double *d,
        const double *sigma,
        const double *rho,
        const double *irho,
        const double *rhoM,
        const double *thetaM,
        double H,
        double Aguide,
        int points,
        const double *xP,
        double *Ra,
        double *Rb,
        double *Rc,
        double *Rd,
    )
    void pnr_MT(
        int workers,
        int layers,
        const double *d,
        const double *sigma,
        const double *rho,
        const double *irho,
        const double *rhoM,
        const double *thetaM,
        double H,
        double Aguide,
        int points,
        const double *xP,
        double *Ra,
        double *Rb,
        double *Rc,
        double *Rd,
    )

ctypedef np.float64_t float64_t


# figure out CPU count
cdef int NCPU = cpu_count()


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef np.ndarray abeles(
    np.ndarray x,
    double[:, :] w,
    double scale=1.0,
    double bkg=0.,
    int threads=-1
):
    """Abeles matrix formalism for calculating reflectivity from a stratified
    medium.

    Parameters
    ----------
    q: array_like
        the q values required for the calculation.
        Q = 4 * Pi / lambda * sin(omega).
        Units = Angstrom**-1
    layers: np.ndarray
        coefficients required for the calculation, has shape (2 + N, 4),
        where N is the number of layers:

        - layers[0, 1]
           SLD of fronting (/1e-6 Angstrom**-2)
        - layers[0, 2]
           iSLD of fronting (/1e-6 Angstrom**-2)
        - layers[N, 0]
           thickness of layer N
        - layers[N, 1]
           SLD of layer N (/1e-6 Angstrom**-2)
        - layers[N, 2]
           iSLD of layer N (/1e-6 Angstrom**-2)
        - layers[N, 3]
           roughness between layer N-1/N
        - layers[-1, 1]
           SLD of backing (/1e-6 Angstrom**-2)
        - layers[-1, 2]
           iSLD of backing (/1e-6 Angstrom**-2)
        - layers[-1, 3]
           roughness between backing and last layer

    scale: float
        Multiply all reflectivities by this value.
    bkg: float
        Linear background to be added to all reflectivities
    threads: int, optional
        How many threads you would like to use in the reflectivity calculation.
        If `threads == -1` then the calculation is automatically spread over
        `multiprocessing.cpu_count()` threads.

    Returns
    -------
    reflectivity: np.ndarray
        Calculated reflectivity values for each q value.
    """
    if w.shape[1] != 4 or w.shape[0] < 2:
        raise ValueError("Layer parameters for _creflect must be an array of"
                         " shape (>2, 4)")
    if x.dtype != np.float64:
        raise ValueError("Q values for _creflect must be np.float64")

    cdef:
        int nlayers = w.shape[0] - 2
        int npoints = x.size
        np.ndarray y = np.empty_like(x, np.float64)
        double *x_data
        double *y_data
    if not x.flags['C_CONTIGUOUS']:
        x = np.ascontiguousarray(x, dtype=np.float64)

    x_data = <float64_t *>np.PyArray_DATA(x)
    y_data = <float64_t *>np.PyArray_DATA(y)

    coefs = <double*> PyMem_Malloc((4*nlayers + 8) * sizeof(double))
    if not coefs:
        raise MemoryError()

    cdef double [:] coefs_view = <double[:4*nlayers + 8]>coefs

    try:
        with nogil:
            if threads == -1:
                threads = NCPU
            elif threads == 0:
                threads = 1

            coefs_view[0] = nlayers
            coefs_view[1] = scale
            coefs_view[2:4] = w[0, 1: 3]
            coefs_view[4: 6] = w[-1, 1: 3]
            coefs_view[6] = bkg
            coefs_view[7] = w[-1, 3]

            if nlayers:
                coefs_view[8::4] = w[1:-1, 0]
                coefs_view[9::4] = w[1:-1, 1]
                coefs_view[10::4] = w[1:-1, 2]
                coefs_view[11::4] = w[1:-1, 3]

            if threads > 1:
                abeles_wrapper_MT(
                    1,
                    4*nlayers + 8,
                    coefs,
                    npoints,
                    y_data,
                    x_data,
                    threads
                )
            else:
                abeles_wrapper(
                    4*nlayers + 8,
                    coefs,
                    npoints,
                    y_data,
                    x_data
                )
    finally:
        PyMem_Free(coefs)

    return y


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef np.ndarray parratt(
    np.ndarray x,
    double[:, :] w,
    double scale=1.0,
    double bkg=0.,
    int threads=-1
):
    """
    Parratt recursion formula for calculating reflectivity from a stratified
    medium.

    Parameters
    ----------
    q: array_like
        the q values required for the calculation.
        Q = 4 * Pi / lambda * sin(omega).
        Units = Angstrom**-1
    layers: np.ndarray
        coefficients required for the calculation, has shape (2 + N, 4),
        where N is the number of layers
        layers[0, 1] - SLD of fronting (/1e-6 Angstrom**-2)
        layers[0, 2] - iSLD of fronting (/1e-6 Angstrom**-2)
        layers[N, 0] - thickness of layer N
        layers[N, 1] - SLD of layer N (/1e-6 Angstrom**-2)
        layers[N, 2] - iSLD of layer N (/1e-6 Angstrom**-2)
        layers[N, 3] - roughness between layer N-1/N
        layers[-1, 1] - SLD of backing (/1e-6 Angstrom**-2)
        layers[-1, 2] - iSLD of backing (/1e-6 Angstrom**-2)
        layers[-1, 3] - roughness between backing and last layer
    scale: float
        Multiply all reflectivities by this value.
    bkg: float
        Linear background to be added to all reflectivities
    threads: int, optional
        How many threads you would like to use in the reflectivity calculation.
        If `threads == -1` then the calculation is automatically spread over
        `multiprocessing.cpu_count()` threads.

    Returns
    -------
    Reflectivity: np.ndarray
        Calculated reflectivity values for each q value.
    """
    if w.shape[1] != 4 or w.shape[0] < 2:
        raise ValueError("Layer parameters for _creflect must be an array of"
                         " shape (>2, 4)")
    if x.dtype != np.float64:
        raise ValueError("Q values for _creflect must be np.float64")

    cdef:
        int nlayers = w.shape[0] - 2
        int npoints = x.size
        np.ndarray y = np.empty_like(x, np.float64)
        double *x_data
        double *y_data
    if not x.flags['C_CONTIGUOUS']:
        x = np.ascontiguousarray(x, dtype=np.float64)

    x_data = <float64_t *>np.PyArray_DATA(x)
    y_data = <float64_t *>np.PyArray_DATA(y)

    coefs = <double*> PyMem_Malloc((4*nlayers + 8) * sizeof(double))
    if not coefs:
        raise MemoryError()

    cdef double [:] coefs_view = <double[:4*nlayers + 8]>coefs

    try:
        with nogil:
            if threads == -1:
                threads = NCPU
            elif threads == 0:
                threads = 1

            coefs_view[0] = nlayers
            coefs_view[1] = scale
            coefs_view[2:4] = w[0, 1: 3]
            coefs_view[4: 6] = w[-1, 1: 3]
            coefs_view[6] = bkg
            coefs_view[7] = w[-1, 3]

            if nlayers:
                coefs_view[8::4] = w[1:-1, 0]
                coefs_view[9::4] = w[1:-1, 1]
                coefs_view[10::4] = w[1:-1, 2]
                coefs_view[11::4] = w[1:-1, 3]

            if threads > 1:
                parratt_wrapper_MT(
                    1,
                    4*nlayers + 8,
                    coefs,
                    npoints,
                    y_data,
                    x_data,
                    threads
                )
            else:
                parratt_wrapper(
                    4*nlayers + 8,
                    coefs,
                    npoints,
                    y_data,
                    x_data
                )
    finally:
        PyMem_Free(coefs)

    return y


@cython.boundscheck(False)
cpdef _contract_by_area(np.ndarray[np.float64_t, ndim=2] slabs, dA=0.5):
    newslabs = np.copy(slabs)[::-1]

    cdef:
        double [:, :] newslabs_view = newslabs
        double [:] d = newslabs_view[:, 0]
        double [:] rho = newslabs_view[:, 1]
        double [:] irho = newslabs_view[:, 2]
        double [:] sigma = newslabs[:, 3]
        double [:] vfsolv = newslabs[:, 4]

        size_t n = np.size(d, 0)
        size_t i, newi
        double dz, rhoarea, irhoarea, vfsolvarea, rholo, rhohi, irholo, irhohi
        double da = float(dA)

    with nogil:
        i = 1
        newi = 1  # skip the substrate

        while i < n:
            # Get ready for the next layer
            # Accumulation of the first row happens in the inner loop
            dz = rhoarea = irhoarea = vfsolvarea = 0.
            rholo = rhohi = rho[i]
            irholo = irhohi = irho[i]

            # Accumulate slices into layer
            while True:
                # Accumulate next slice
                dz += d[i]
                rhoarea += d[i] * rho[i]
                irhoarea += d[i] * irho[i]
                vfsolvarea += d[i] * vfsolv[i]

                i += 1
                # If no more slices or sigma != 0, break immediately
                if i == n or sigma[i - 1] != 0.:
                    break

                # If next slice won't fit, break
                if rho[i] < rholo:
                    rholo = rho[i]
                if rho[i] > rhohi:
                    rhohi = rho[i]
                if (rhohi - rholo) * (dz + d[i]) > da:
                    break

                if irho[i] < irholo:
                    irholo = irho[i]
                if irho[i] > irhohi:
                    irhohi = irho[i]
                if (irhohi - irholo) * (dz + d[i]) > da:
                    break

            # Save the layer
            d[newi] = dz
            if i == n:
                # printf("contract: adding final sld at %d\n",newi)
                # Last layer uses surface values
                rho[newi] = rho[n - 1]
                irho[newi] = irho[n - 1]
                vfsolv[newi] = vfsolv[n - 1]
            else:
                # Middle layers uses average values
                rho[newi] = rhoarea / dz
                irho[newi] = irhoarea / dz
                sigma[newi] = sigma[i - 1]
                vfsolv[newi] = vfsolvarea / dz
            # First layer uses substrate values
            newi += 1

    return newslabs[:newi][::-1]


cpdef np.ndarray abeles_vectorised(
    np.ndarray x,
    double[:, :, :] w,
    scale=None,
    bkg=None,
    int threads=-1
):
    """
    Vectorised Abeles matrix formalism for calculating reflectivity from a
    stratified medium.

    Parameters
    ----------
    q: array_like
        the q values required for the calculation.
        Q = 4 * Pi / lambda * sin(omega).
        Units = Angstrom**-1
    layers: np.ndarray
        coefficients required for the calculation, has shape (M, 2 + N, 4).
        The calculation is vectorised over the M sets of film parameters, and
        N is the number of layers in each film.

        - layers[0, 1]
           SLD of fronting (/1e-6 Angstrom**-2)
        - layers[0, 2]
           iSLD of fronting (/1e-6 Angstrom**-2)
        - layers[N, 0]
           thickness of layer N
        - layers[N, 1]
           SLD of layer N (/1e-6 Angstrom**-2)
        - layers[N, 2]
           iSLD of layer N (/1e-6 Angstrom**-2)
        - layers[N, 3]
           roughness between layer N-1/N
        - layers[-1, 1]
           SLD of backing (/1e-6 Angstrom**-2)
        - layers[-1, 2]
           iSLD of backing (/1e-6 Angstrom**-2)
        - layers[-1, 3]
           roughness between backing and last layer

    scale: array-like, optional
        Multiply all reflectivities by this value.
    bkg: array-like, optional
        Linear background to be added to all reflectivities
    threads: int, optional
        How many threads you would like to use in the reflectivity calculation.
        If `threads == -1` then the calculation is automatically spread over
        `multiprocessing.cpu_count()` threads.

    Returns
    -------
    Reflectivity: np.ndarray
        Calculated reflectivity values for each q value.
    """
    if w.shape[2] != 4 or w.shape[1] < 2:
        raise ValueError(
            "Layer parameters for _creflect.abeles_vectorised must be an"
            "array of shape (>=1, >2, 4)"
        )
    if x.dtype != np.float64:
        raise ValueError("Q values for _creflect.abeles_vectorised must be np.float64")

    if scale is not None:
        if not isinstance(scale, np.ndarray) or scale.shape[0] != w.shape[0]:
            raise ValueError("scale must be an array of shape (M,)")
    else:
        scale = np.ones(w.shape[0], dtype=np.float64)

    if bkg is not None:
        if not isinstance(bkg, np.ndarray) or bkg.shape[0] != w.shape[0]:
            raise ValueError("bkg must be an array of shape (M,)")
    else:
        bkg = np.zeros(w.shape[0])

    if threads == -1:
        threads = NCPU
    elif threads == 0:
        threads = 1

    cdef:
        int nlayers = w.shape[1] - 2
        int i
        int j
        int offset
        unsigned int batch = w.shape[0]
        int npoints = x.size
        np.ndarray yout = np.repeat(
            np.empty_like(x, np.float64)[np.newaxis, ...], batch, axis=0
        )
        double *x_data
        double *bkg_data
        double *scale_data
        double *y_out_data
        double *coefs
        double *coefs_arr
    if not x.flags['C_CONTIGUOUS']:
        x = np.ascontiguousarray(x, dtype=np.float64)

    x_data = <float64_t *>np.PyArray_DATA(x)
    y_out_data = <float64_t *>np.PyArray_DATA(yout)
    scale_data = <float64_t *>np.PyArray_DATA(scale)
    bkg_data = <float64_t *> np.PyArray_DATA(bkg)

    offset = 4*nlayers + 8
    coefs_arr = <double*> PyMem_Malloc(offset * sizeof(double) * batch)
    if not coefs_arr:
        raise MemoryError()

    try:
        for i in range(batch):
            _o = offset * i
            coefs = &coefs_arr[_o]
            coefs[0] = nlayers
            coefs[1] = scale_data[i]
            coefs[2] = w[i, 0, 1]
            coefs[3] = w[i, 0, 2]
            coefs[4] = w[i, -1, 1]
            coefs[5] = w[i, -1, 2]
            coefs[6] = bkg_data[i]
            coefs[7] = w[i, -1, 3]

            if nlayers:
                for j in range(nlayers):
                    coefs[8 + 4*j] = w[i, j + 1, 0]
                    coefs[9 + 4*j] = w[i, j + 1, 1]
                    coefs[10 + 4*j] = w[i, j + 1, 2]
                    coefs[11 + 4*j] = w[i, j + 1, 3]

        abeles_wrapper_MT(
            batch,
            4*nlayers + 8,
            coefs_arr,
            npoints,
            y_out_data,
            x_data,
            threads
        )
    finally:
        PyMem_Free(coefs_arr)

    return yout


cpdef np.ndarray gepore(
    np.ndarray x,
    double[:, :] w,
    double H=0.0,
    double Aguide=270,
    double scale=1.0,
    double bkg=0.,
    int threads=-1,
):
    """
    Abeles matrix formalism for calculating polarised neutron reflectivity
    from a stratified medium.

    Parameters
    ----------
    q: array_like
        the q values required for the calculation.
        Q = 4 * Pi / lambda * sin(omega).
        Units = Angstrom**-1
    layers: np.ndarray
        coefficients required for the calculation, has shape (2 + N, 6),
        where N is the number of layers
        layers[0, 1] - SLD of fronting (/1e-6 Angstrom**-2)
        layers[0, 2] - iSLD of fronting (/1e-6 Angstrom**-2)
        layers[N, 0] - thickness of layer N
        layers[N, 1] - SLD of layer N (/1e-6 Angstrom**-2)
        layers[N, 2] - iSLD of layer N (/1e-6 Angstrom**-2)
        layers[N, 3] - roughness between layer N-1/N
        layers[N, 4] - magnetic SLD (/1e-6 Angstrom**-2)
        layers[N, 5] - angle of the magnetism within the layer (degrees)
        layers[-1, 1] - SLD of backing (/1e-6 Angstrom**-2)
        layers[-1, 2] - iSLD of backing (/1e-6 Angstrom**-2)
        layers[-1, 3] - roughness between backing and last layer
    H: float
        Applied magnetic field (T)
    Aguide: float
        Orientation angle for applied magnetic field (degrees). Should be 270
        or 90 for field in plane of sample.
    scale: float
        Multiply all reflectivities by this value.
    bkg: float
        Linear background to be added to all reflectivities
    threads: int, optional
        How many threads you would like to use in the reflectivity calculation.
        If `threads == -1` then the calculation is automatically spread over
        `multiprocessing.cpu_count()` threads.

    Returns
    -------
    Reflectivity: np.ndarray
        Calculated reflectivity values for each q value. Has shape `(4, M)`
        where M is the number of points in `q`.
    """
    if w.shape[1] != 6 or w.shape[0] < 2:
        raise ValueError(
            "Layer parameters for _creflect.gepore must be an"
            "array of shape (>=2, 6)"
        )
    if x.dtype != np.float64:
        raise ValueError("Q values for _creflect.gepore must be np.float64")
    if not x.flags['C_CONTIGUOUS']:
        x = np.ascontiguousarray(x, dtype=np.float64)

    cdef:
        int npoints = x.size
        np.ndarray y = np.zeros((4, npoints), np.float64)
        const double *xP
        double *d_data
        double *sigma_data
        double *rho_data
        double *irho_data
        double *rhoM_data
        double *thetaM_data

        double *Ra
        double *Rb
        double *Rc
        double *Rd
        int layers = w.shape[0]
        d = np.zeros(layers)
        sigma = np.zeros(layers - 1)
        rho = np.zeros(layers)
        irho = np.zeros(layers)
        rhoM = np.zeros(layers)
        thetaM = np.zeros(layers)

    if threads == -1:
        threads = NCPU
    elif threads == 0:
        threads = 1

    xP = <float64_t *> np.PyArray_DATA(x)

    d[:] = w[:, 0]
    sigma[:] = w[1:, 3]
    rho[:] = w[:, 1]
    irho[:] = w[:, 2]
    rhoM[:] = w[:, 4]
    thetaM[:] = w[:, 5]

    d_data = <float64_t *>np.PyArray_DATA(d)
    sigma_data = <float64_t *>np.PyArray_DATA(sigma)
    rho_data = <float64_t *>np.PyArray_DATA(rho)
    irho_data = <float64_t *>np.PyArray_DATA(irho)
    rhoM_data = <float64_t *>np.PyArray_DATA(rhoM)
    thetaM_data = <float64_t *>np.PyArray_DATA(thetaM)

    Ra = <float64_t *> np.PyArray_DATA(y[0])
    Rb = <float64_t *> np.PyArray_DATA(y[1])
    Rc = <float64_t *> np.PyArray_DATA(y[2])
    Rd = <float64_t *> np.PyArray_DATA(y[3])

    if threads > 1:
        pnr_MT(
            threads,
            layers,
            d_data,
            sigma_data,
            rho_data,
            irho_data,
            rhoM_data,
            thetaM_data,
            H,
            Aguide,
            npoints,
            xP,
            Ra,
            Rb,
            Rc,
            Rd
        )
    else:
        pnr(
            layers,
            d_data,
            sigma_data,
            rho_data,
            irho_data,
            rhoM_data,
            thetaM_data,
            H,
            Aguide,
            npoints,
            xP,
            Ra,
            Rb,
            Rc,
            Rd
        )
    y *= scale
    y += bkg
    return y



"""
An experimental simulator for a TOF neutron reflectometer
"""

__author__ = "Andrew Nelson"
__copyright__ = "Copyright 2019, Andrew Nelson"
__license__ = "3 clause BSD"

import numpy as np
from scipy.integrate import simpson
from scipy.interpolate import InterpolatedUnivariateSpline as IUS
from scipy.stats import rv_continuous, trapezoid, norm, uniform
from scipy.optimize import brentq
from scipy._lib._util import check_random_state

from refnx.reduce.platypusnexus import (
    calculate_wavelength_bins,
    create_reflect_nexus,
    PlatypusNexus,
)
from refnx.reduce import parabolic_motion as pm
from refnx.util import general, ErrorProp
from refnx.util._resolution_kernel import AngularDivergence
from refnx.reflect import Slab, Structure, SLD, ReflectModel
from refnx.dataset import ReflectDataset


class SpectrumDist(rv_continuous):
    """
    The `SpectrumDist` object is a `scipy.stats` like object to describe the
    neutron intensity as a function of wavelength. You can use the `pdf, cdf,
    ppf, rvs` methods like you would a `scipy.stats` distribution. Of
    particular interest is the `rvs` method which randomly samples neutrons
    whose distribution obeys the direct beam spectrum. Random variates are
    generated the `rv_continuous` superclass by classical generation of
    uniform noise coupled with the `ppf`. `ppf` is approximated by linear
    interpolation of `q` into a pre-calculated inverse `cdf`.
    """

    def __init__(self, x, y):
        super(SpectrumDist, self).__init__(a=np.min(x), b=np.max(x))
        self._x = x

        # normalise the distribution
        area = simpson(y, x=x)
        y /= area
        self._y = y

        # an InterpolatedUnivariate spline models the spectrum
        self.spl = IUS(x, y)

        # fudge_factor required because integral of the spline is not exactly 1
        self.fudge_factor = self.spl.integral(self.a, self.b)

        # calculate a gridded and sampled version of the CDF.
        # this can be used with interpolation for quick calculation
        # of ppf (necessary for quick rvs)
        self._x_interpolated_cdf = np.linspace(np.min(x), np.max(x), 1000)
        self._interpolated_cdf = self.cdf(self._x_interpolated_cdf)

    def _pdf(self, x):
        return self.spl(x) / self.fudge_factor

    def _cdf(self, x):
        xflat = x.ravel()

        def f(x):
            return self.spl.integral(self.a, x) / self.fudge_factor

        v = map(f, xflat)

        r = np.fromiter(v, dtype=float).reshape(x.shape)
        return r

    def _f(self, x, qq):
        return self._cdf(x) - qq

    def _g(self, qq, *args):
        return brentq(self._f, self._a, self._b, args=(qq,) + args)

    def _ppf(self, q, *args):
        qflat = q.ravel()
        """
        _a, _b = self._get_support(*args)

        def f(x, qq):
            return self._cdf(x) - qq

        def g(qq):
            return brentq(f, _a, _b, args=(qq,) + args, xtol=1e-3)

        v = map(g, qflat)

        cdf = _CDF(self.spl, self.fudge_factor, _a, _b)
        g = _G(cdf)

        with Pool() as p:
            v = p.map(g, qflat)
            r = np.fromiter(v, dtype=float).reshape(q.shape)
        """
        # approximate the ppf using a sampled+interpolated CDF
        # the commented out methods are more accurate, but are at least
        # 3 orders of magnitude slower.
        r = np.interp(qflat, self._interpolated_cdf, self._x_interpolated_cdf)
        return r.reshape(q.shape)


# for parallelisation (can't pickle rv_continuous all that easily)
class _CDF(object):
    def __init__(self, spl, fudge_factor, a, b):
        self.a = a
        self.b = b
        self.spl = spl
        self.fudge_factor = fudge_factor

    def __call__(self, x):
        return self.spl.integral(self.a, x) / self.fudge_factor


class _G(object):
    def __init__(self, cdf):
        self.cdf = cdf

    def _f(self, x, qq):
        return self.cdf(x) - qq

    def __call__(self, q):
        return brentq(self._f, self.cdf.a, self.cdf.b, args=(q,), xtol=1e-4)


class ReflectSimulator(object):
    """
    Simulate a reflectivity pattern from PLATYPUS.

    Parameters
    ----------
    model : refnx.reflect.ReflectModel

    angle : float
        Angle of incidence (degrees)

    p_theta : AngularDivergence

    L2S : float
        Distance from pre-sample slit to sample

    direct_spectrum : str, ReflectNexus, h5 handle
        Contains the direct beam spectrum. Processed using ReflectNexus

    lo_wavelength : float
        smallest wavelength used from the generated neutron spectrum

    hi_wavelength : float
        longest wavelength used from the generated neutron spectrum

    dlambda : float
        Wavelength resolution expressed as a percentage. dlambda=3.3
        corresponds to using disk choppers 1+3 on *PLATYPUS*.
        (FWHM of the Gaussian approximation of a trapezoid)

    rebin : float
        Rebinning expressed as a percentage. The width of a wavelength bin is
        `rebin / 100 * lambda`. You have to multiply by 0.68 to get its
        fractional contribution to the overall resolution smearing.

    gravity : bool
        Apply gravity during simulation? Turn gravity off if `angle` is
        already an array of incident angles due to gravity.

    force_gaussian : bool
        Instead of using trapzeoidal and uniform distributions for angular
        and wavelength resolution, use a Gaussian distribution (doesn't apply
        to the rebinning contribution).

    force_uniform_wavelength : bool
        Instead of using a wavelength spectrum representative of a
        time-of-flight reflectometer generate wavelengths from a uniform
        distribution.

    only_resolution : bool
        Set to `True` if this class is only going to be used to calculate
        detailed resolution kernels.

    Notes
    -----
    Angular, chopper and rebin smearing effects are all taken into account.
    """

    def __init__(
        self,
        model,
        angle,
        p_theta,
        L2S=120.0,
        direct_spectrum=None,
        lo_wavelength=2.8,
        hi_wavelength=18,
        dlambda=3.3,
        rebin=2,
        gravity=False,
        force_gaussian=False,
        force_uniform_wavelength=False,
        only_resolution=False,
    ):
        self.model = model

        self.bkg = model.bkg.value
        self.angle = angle
        self.p_theta = p_theta
        assert isinstance(p_theta, AngularDivergence)

        # dlambda refers to the FWHM of the gaussian approximation to a uniform
        # distribution. The full width of the uniform distribution is
        # dlambda/0.68.
        self.dlambda = dlambda / 100.0
        # the rebin percentage refers to the full width of the bins. You have to
        # multiply this value by 0.68 to get the equivalent contribution to the
        # resolution function.
        self.rebin = rebin / 100.0
        self.wavelength_bins = calculate_wavelength_bins(
            lo_wavelength, hi_wavelength, rebin
        )
        bin_centre = 0.5 * (
            self.wavelength_bins[1:] + self.wavelength_bins[:-1]
        )

        # angular deviation due to gravity
        # --> no correction for gravity affecting width of angular resolution
        elevations = 0
        if gravity:
            speeds = general.wavelength_velocity(bin_centre)
            # trajectories through slits for different wavelengths
            trajectories = pm.find_trajectory(p_theta.L12 / 1000.0, 0, speeds)
            # elevation at sample
            elevations = pm.elevation(
                trajectories, speeds, (p_theta.L12 + L2S) / 1000.0
            )

        # nominal Q values
        self.q = general.q(angle - elevations, bin_centre)

        # keep a tally of the direct and reflected beam
        self.direct_beam = np.zeros((self.wavelength_bins.size - 1))
        self.reflected_beam = np.zeros((self.wavelength_bins.size - 1))

        # beam monitor counts for normalisation
        self.bmon_direct = 0
        self.bmon_reflect = 0

        self.gravity = gravity

        # wavelength generator
        self.force_uniform_wavelength = force_uniform_wavelength
        if force_uniform_wavelength:
            self.spectrum_dist = uniform(
                loc=lo_wavelength - 1, scale=hi_wavelength - lo_wavelength + 1
            )
        else:
            a = create_reflect_nexus(direct_spectrum)
            direct = False
            if isinstance(a, PlatypusNexus):
                direct = True

            q, i, di = a.process(
                normalise=False,
                normalise_bins=False,
                rebin_percent=0.5,
                lo_wavelength=max(0, lo_wavelength - 1),
                hi_wavelength=hi_wavelength + 1,
                direct=direct,
            )
            q = q.squeeze()
            i = i.squeeze()
            self.spectrum_dist = SpectrumDist(q, i)

        self.force_gaussian = force_gaussian
        self.only_resolution = only_resolution

        # angular resolution generator, based on a trapezoidal distribution
        # The slit settings are the optimised set typically used in an
        # experiment. dtheta/theta refers to the FWHM of a Gaussian
        # approximation to a trapezoid.

        # stores the q vectors contributing towards each datapoint
        self._res_kernel = {}
        self._min_samples = 0

        self.angle = angle
        self.L12 = p_theta.L12
        self.L2S = L2S
        self.div = p_theta.dtheta
        self.s1 = p_theta.d1
        self.s2 = p_theta.d2

        if force_gaussian:
            self.angular_dist = norm(scale=self.div / 2.3548)
        else:
            self.angular_dist = p_theta.rv

    def sample(self, samples, random_state=None):
        """
        Sample the beam for reflected signal.

        2400000 samples roughly corresponds to 1200 sec of *PLATYPUS* using
        dlambda=3.3 and dtheta=3.3 at angle=0.65.
        150000000 samples roughly corresponds to 3600 sec of *PLATYPUS* using
        dlambda=3.3 and dtheta=3.3 at angle=3.0.

        (The sample number <--> actual acquisition time correspondence has
         not been checked fully)

        Parameters
        ----------
        samples: int
            How many samples to run.
        random_state: {int, `~np.random.RandomState`, `~np.random.Generator`}, optional
            If `random_state` is not specified the
            `~np.random.RandomState` singleton is used.
            If `random_state` is an int, a new ``RandomState`` instance is used,
            seeded with seed.
            If `random_state` is already a ``RandomState`` or a ``Generator``
            instance, then that object is used.
            Specify `random_state` for repeatable minimizations.
        """
        # grab a random number generator
        rng = check_random_state(random_state)

        # generate neutrons of various wavelengths
        wavelengths = self.spectrum_dist.rvs(size=samples, random_state=rng)

        # generate neutrons of different angular divergence
        angles = self.angular_dist.rvs(samples, random_state=rng) + self.angle

        # angular deviation due to gravity
        # --> no correction for gravity affecting width of angular resolution
        if self.gravity:
            speeds = general.wavelength_velocity(wavelengths)
            # trajectories through slits for different wavelengths
            trajectories = pm.find_trajectory(self.L12 / 1000.0, 0, speeds)
            # elevation at sample
            elevations = pm.elevation(
                trajectories, speeds, (self.L12 + self.L2S) / 1000.0
            )
            angles -= elevations

        # calculate Q
        q = general.q(angles, wavelengths)

        # calculate reflectivities for a neutron of a given Q.
        # the angular resolution smearing has already been done. The wavelength
        # resolution smearing follows.
        if not self.only_resolution:
            r = self.model(q, x_err=0.0)

            # accept or reject neutrons based on the reflectivity of
            # sample at a given Q.
            criterion = rng.uniform(size=samples)
            accepted = criterion < r
        else:
            accepted = np.ones_like(q, dtype=bool)

        # implement wavelength smearing from choppers. Jitter the wavelengths
        # by a uniform distribution whose full width is dlambda / 0.68.
        if self.force_gaussian:
            noise = rng.standard_normal(size=samples)
            jittered_wavelengths = wavelengths * (
                1 + self.dlambda / 2.3548 * noise
            )
        else:
            noise = rng.uniform(-0.5, 0.5, size=samples)
            jittered_wavelengths = wavelengths * (
                1 + self.dlambda / 0.68 * noise
            )

        # update reflected beam counts. Rebin smearing
        # is taken into account due to the finite size of the wavelength
        # bins.
        hist = np.histogram(
            jittered_wavelengths[accepted], self.wavelength_bins
        )
        self.reflected_beam += hist[0]
        self.bmon_reflect += float(samples)

        # update resolution kernel. If we have more than 100000 in all
        # bins skip
        if (
            len(self._res_kernel)
            and np.min([len(v) for v in self._res_kernel.values()]) > 1000000
        ):
            return

        bin_loc = np.digitize(jittered_wavelengths, self.wavelength_bins)
        for i in range(1, len(self.wavelength_bins)):
            # extract q values that fall in each wavelength bin
            q_for_bin = np.copy(q[bin_loc == i])
            q_samples_so_far = self._res_kernel.get(i - 1, np.array([]))
            updated_samples = np.concatenate((q_samples_so_far, q_for_bin))

            # no need to keep double precision for these sample arrays
            self._res_kernel[i - 1] = updated_samples.astype(np.float32)

    def sample_direct(self, samples, random_state=None):
        """
        Samples the direct beam

        Parameters
        ----------
        samples: int
            How many samples to run.
        random_state: {int, `~np.random.RandomState`, `~np.random.Generator`}, optional
            If `random_state` is not specified the
            `~np.random.RandomState` singleton is used.
            If `random_state` is an int, a new ``RandomState`` instance is used,
            seeded with seed.
            If `random_state` is already a ``RandomState`` or a ``Generator``
            instance, then that object is used.
            Specify `random_state` for repeatable minimizations.
        """
        # grab a random number generator
        rng = check_random_state(random_state)

        # generate neutrons of various wavelengths
        wavelengths = self.spectrum_dist.rvs(size=samples, random_state=rng)

        # implement wavelength smearing from choppers. Jitter the wavelengths
        # by a uniform distribution whose full width is dlambda / 0.68.
        if self.force_gaussian:
            noise = rng.standard_normal(size=samples)
            jittered_wavelengths = wavelengths * (
                1 + self.dlambda / 2.3548 * noise
            )
        else:
            noise = rng.uniform(-0.5, 0.5, size=samples)
            jittered_wavelengths = wavelengths * (
                1 + self.dlambda / 0.68 * noise
            )

        hist = np.histogram(jittered_wavelengths, self.wavelength_bins)

        self.direct_beam += hist[0]
        self.bmon_direct += float(samples)

    @property
    def resolution_kernel(self):
        histos = []
        # first histogram all the q values corresponding to a specific bin
        # this will come as shortest wavelength first, or highest Q. This
        # is because the wavelength bins are monotonic increasing.
        for v in self._res_kernel.values():
            histos.append(np.histogram(v, density=True, bins="auto"))

        # make lowest Q comes first.
        histos.reverse()

        # what is the largest number of bins?
        max_bins = np.max([len(histo[0]) for histo in histos])

        kernel = np.full((len(histos), 2, max_bins), np.nan)
        for i, histo in enumerate(histos):
            p, x = histo
            sz = len(p)
            kernel[i, 0, :sz] = 0.5 * (x[:-1] + x[1:])
            kernel[i, 1, :sz] = p

        # filter points with zero counts because error is incorrect
        mask = self.reflected_beam != 0
        mask = mask[::-1]

        kernel = kernel[mask]

        return kernel

    @property
    def reflectivity(self):
        """
        The reflectivity of the sampled system
        """
        if self.only_resolution:
            raise RuntimeError(
                "ReflectSimulator is only calculating resolution function"
            )

        rerr = np.sqrt(self.reflected_beam)
        bmon_reflect_err = np.sqrt(self.bmon_reflect)

        ierr = np.sqrt(self.direct_beam)
        bmon_direct_err = np.sqrt(self.bmon_direct)

        # reverse the resolution kernel, because that's output in sorted order
        dx = self.resolution_kernel[::-1]

        # divide reflectivity signal by bmon
        ref, rerr = ErrorProp.EPdiv(
            self.reflected_beam, rerr, self.bmon_reflect, bmon_reflect_err
        )
        # divide direct signal by bmon
        direct, ierr = ErrorProp.EPdiv(
            self.direct_beam, ierr, self.bmon_direct, bmon_direct_err
        )

        # now calculate reflectivity
        ref, rerr = ErrorProp.EPdiv(ref, rerr, direct, ierr)

        # filter points with zero counts because error is incorrect
        mask = rerr != 0

        dataset = ReflectDataset(
            data=(self.q[mask], ref[mask], rerr[mask], dx)
        )

        dataset.sort()
        # apply some counting statistics on top of dataset otherwise there will
        # be no variation at e.g. critical edge.
        # return dataset.synthesise()
        return dataset



import pytest
import refnx.reduce.xray as xray


def test_reduction_runs(data_directory):
    # just ensure that the reduction occurs without raising an Exception.
    # We're not testing for correctness here (yet)
    fpath = data_directory / "reduce" / "180706_HA_DG2.xrdml"
    xray.reduce_xrdml(fpath)



# cython: language_level=3, boundscheck=False

from libc.stdlib cimport abort, malloc, free
import numpy as np
cimport numpy as np
cimport cython
cimport openmp
from cython.parallel import prange, parallel, threadid


np.import_array()
DTYPE = np.float64
ctypedef np.float64_t DTYPE_t
TINY = 1e-30


cdef extern from "<complex>" namespace "std" nogil:
    double complex sqrt(double complex)
    double complex exp(double complex)
    double complex conj(double complex)


cdef extern from "<math.h>" nogil:
    double fabs(double)


cdef extern from "refcaller.h" nogil:
    void abeles_wrapper(int numcoefs, const double *coefP, int npoints, double *yP,
                 const double *xP)


cpdef abeles(x, np.ndarray[DTYPE_t, ndim=2] w,
             double scale=1.0, double bkg=0., int threads=-1):

    # we need the abscissae in a contiguous block of memory
    cdef double[:] xtemp = np.ascontiguousarray(x, dtype=DTYPE).flatten()

    if w.shape[1] != 4 or w.shape[0] < 2:
        raise ValueError("Parameters for reflection must have shape (>2, 4)")

    if threads > 0:
        num_threads = int(threads)
    else:
        num_threads = openmp.omp_get_max_threads()

    y = _abeles(xtemp, w, scale, bkg, num_threads)

    y = np.reshape(y, x.shape)
    return y


cpdef parratt(x, np.ndarray[DTYPE_t, ndim=2] w,
             double scale=1.0, double bkg=0., int threads=-1):

    # we need the abscissae in a contiguous block of memory
    cdef double[:] xtemp = np.ascontiguousarray(x, dtype=DTYPE).flatten()

    if w.shape[1] != 4 or w.shape[0] < 2:
        raise ValueError("Parameters for reflection must have shape (>2, 4)")

    if threads > 0:
        num_threads = int(threads)
    else:
        num_threads = openmp.omp_get_max_threads()

    y = _parratt(xtemp, w, scale, bkg, num_threads)

    y = np.reshape(y, x.shape)
    return y


@cython.boundscheck(False)
@cython.cdivision(True)
cdef _abeles(double[:] x,
             np.ndarray[np.float64_t, ndim=2] w,
             double scale=1.0, double bkg=0., int num_threads=1):

    cdef:
        int _num_threads = num_threads
        int nlayers = w.shape[0] - 2
        int npoints = x.shape[0]
        int layer, i, m
        double complex M_4PI = 4 * np.pi
        double complex I = 1j
        double complex rj, k, k_next, q2, rough, mi00, mi01, mi10, mi11, thick
        double complex mrtot00, mrtot01, mrtot10, mrtot11, p0, p1, beta, arg

        np.ndarray[np.complex128_t, ndim=1] y = np.zeros(npoints,
                                                         np.complex128)
        np.ndarray[np.complex128_t, ndim=1] roughsqr = np.empty(
            nlayers + 1, np.complex128)

        np.ndarray[np.complex128_t, ndim=1] SLD = np.zeros(
            (w.shape[0]), np.complex128)

        double[:, :] wbuf = w
        double complex[:] SLDbuf = SLD
        double complex[:] roughbuf = roughsqr

    for i in range(1, nlayers + 2):
        SLDbuf[i] = M_4PI * (wbuf[i, 1] - wbuf[0, 1] +
                             1j * (fabs(wbuf[i, 2]) + TINY)) * 1.e-6
        roughbuf[i - 1] = -2. * wbuf[i, 3] * wbuf[i, 3]

    for i in prange(npoints, nogil=True, num_threads=_num_threads):
        q2 = x[i] * x[i] / 4.
        k = x[i] / 2.
        for m in range(0, nlayers + 1):
            k_next = sqrt(q2 - SLDbuf[m + 1])
            rj = (k - k_next) / (k + k_next) * exp(k * k_next * roughbuf[m])

            if not m:
                # characteristic matrix for first interface
                mrtot00 = 1.
                mrtot01 = rj
                mrtot11 = 1.
                mrtot10 = rj
            else:
                # work out the beta for the layer
                thick = wbuf[m, 0]
                beta = exp(k * thick * I)
                # this is the characteristic matrix of a layer
                mi00 = beta
                mi11 = 1. / beta
                mi10 = rj * mi00
                mi01 = rj * mi11

                # matrix multiply
                p0 = mrtot00 * mi00 + mrtot10 * mi01
                p1 = mrtot00 * mi10 + mrtot10 * mi11
                mrtot00 = p0
                mrtot10 = p1

                p0 = mrtot01 * mi00 + mrtot11 * mi01
                p1 = mrtot01 * mi10 + mrtot11 * mi11

                mrtot01 = p0
                mrtot11 = p1

            k = k_next

        y[i] = (mrtot01 / mrtot00)
        y[i] = y[i] * conj(y[i])
        y[i] *= scale
        y[i] += bkg
    return y.real


@cython.boundscheck(False)
@cython.cdivision(True)
cdef _parratt(double[:] x,
              np.ndarray[np.float64_t, ndim=2] w,
              double scale=1.0, double bkg=0., int num_threads=1):

    cdef:
        int _num_threads = num_threads
        int nlayers = w.shape[0] - 2
        int npoints = x.shape[0]
        int layer, i, idx
        double complex M_4PI = 4 * np.pi
        double complex I = 1j
        double complex rj, kn, kn_next, qq2, rough
        double complex beta, arg, RRJ, RRJ_1

        np.ndarray[np.complex128_t, ndim=1] y = np.zeros(npoints,
                                                         np.complex128)
        np.ndarray[np.complex128_t, ndim=1] roughsqr = np.empty(
            nlayers + 1, np.complex128)

        np.ndarray[np.complex128_t, ndim=1] SLD = np.zeros(
            (w.shape[0]), np.complex128)

        double[:, :] wbuf = w
        double complex[:] SLDbuf = SLD
        double complex[:] roughbuf = roughsqr

    for idx in range(1, nlayers + 2):
        SLDbuf[idx] = M_4PI * (wbuf[idx, 1] - wbuf[0, 1] +
                               I * (fabs(wbuf[idx, 2]) + TINY)) * 1.e-6
        roughbuf[idx - 1] = -2. * wbuf[idx, 3] * wbuf[idx, 3]

    for i in prange(npoints, nogil=True, num_threads=_num_threads):
        qq2 = x[i] * x[i] / 4.

        # start from subphase
        kn_next = sqrt(qq2 - SLD[nlayers + 1])

        for idx in range(nlayers, -1, -1):
            # wavevector in the layer
            kn = sqrt(qq2 - SLD[idx])

            # reflectance of the interface
            # factor of 2 is already incorporated in rough_sqr
            rj = ((kn - kn_next)/(kn + kn_next)
                  * exp(kn * kn_next * roughbuf[idx]))

            if (idx == nlayers):
                # characteristic matrix for first interface
                RRJ = rj
            else:
                beta = exp(-2.0 * I * kn_next * fabs(wbuf[idx + 1, 0]))
                RRJ = (rj + RRJ_1 * beta) / (1 + RRJ_1 * beta * rj)

            kn_next = kn
            RRJ_1 = RRJ

        y[i] = RRJ_1 * conj(RRJ_1)
        y[i] *= scale
        y[i] += bkg
    return y.real


cpdef np.ndarray abeles_vectorised(
    np.ndarray x,
    double[:, :, :] w,
    scale=None,
    bkg=None,
    int threads=-1):
    """
    Vectorised Abeles matrix formalism for calculating reflectivity from a
    stratified medium.

    Parameters
    ----------
    q: array_like
        the q values required for the calculation.
        Q = 4 * Pi / lambda * sin(omega).
        Units = Angstrom**-1
    layers: np.ndarray
        coefficients required for the calculation, has shape (M, 2 + N, 4).
        The calculation is vectorised over the M sets of film parameters, and
        N is the number of layers in each film.
        layers[:, 0, 1] - SLD of fronting (/1e-6 Angstrom**-2)
        layers[:, 0, 2] - iSLD of fronting (/1e-6 Angstrom**-2)
        layers[:, N, 0] - thickness of layer N
        layers[:, N, 1] - SLD of layer N (/1e-6 Angstrom**-2)
        layers[:, N, 2] - iSLD of layer N (/1e-6 Angstrom**-2)
        layers[:, N, 3] - roughness between layer N-1/N
        layers[:, -1, 1] - SLD of backing (/1e-6 Angstrom**-2)
        layers[:, -1, 2] - iSLD of backing (/1e-6 Angstrom**-2)
        layers[:, -1, 3] - roughness between backing and last layer
    scale: array-like, optional
        Multiply all reflectivities by this value.
    bkg: array-like, optional
        Linear background to be added to all reflectivities
    threads: int, optional
        How many threads you would like to use in the reflectivity calculation.
        If `threads == -1` then the calculation is automatically spread over
        `multiprocessing.cpu_count()` threads.

    Returns
    -------
    Reflectivity: np.ndarray
        Calculated reflectivity values for each q value.
    """
    if w.shape[2] != 4 or w.shape[1] < 2:
        raise ValueError(
            "Layer parameters for _creflect.vec_abeles must be an"
            "array of shape (>=1, >2, 4)"
        )
    if x.dtype != np.float64:
        raise ValueError("Q values for _creflect.vec_abeles must be np.float64")

    if scale is not None:
        if not isinstance(scale, np.ndarray) or scale.shape != w.shape[0]:
            raise ValueError("scale must be an array of shape (M,)")
    else:
        scale = np.ones(w.shape[0], dtype=np.float64)

    if bkg is not None:
        if not isinstance(bkg, np.ndarray) or bkg.shape != w.shape[0]:
            raise ValueError("bkg must be an array of shape (M,)")
    else:
        bkg = np.zeros(w.shape[0])

    cdef:
        int nlayers = w.shape[1] - 2
        int i
        int j
        int num_threads
        int offset
        int nvec = w.shape[0]
        int npoints = x.size
        np.ndarray yout = np.repeat(
            np.empty_like(x, np.float64)[np.newaxis, ...], nvec, axis=0
        )
        double *x_data
        double *bkg_data
        double *scale_data
        double *y_out_data
        double *coefs
        double *coefs_arr
    if not x.flags['C_CONTIGUOUS']:
        x = np.ascontiguousarray(x, dtype=np.float64)

    x_data = <DTYPE_t *>np.PyArray_DATA(x)
    y_out_data = <DTYPE_t *>np.PyArray_DATA(yout)
    scale_data = <DTYPE_t *>np.PyArray_DATA(scale)
    bkg_data = <DTYPE_t *> np.PyArray_DATA(bkg)

    if threads > 0:
        num_threads = int(threads)
    else:
        num_threads = openmp.omp_get_max_threads()

    offset = 4*nlayers + 8
    coefs_arr = <double *> malloc(offset * sizeof(double) * num_threads)
    if coefs_arr is NULL:
        abort()

    for i in prange(nvec, nogil=True, num_threads=num_threads):
        coefs = &coefs_arr[offset * threadid()]
        coefs[0] = nlayers
        coefs[1] = scale_data[i]
        coefs[2] = w[i, 0, 1]
        coefs[3] = w[i, 0, 2]
        coefs[4] = w[i, -1, 1]
        coefs[5] = w[i, -1, 2]
        coefs[6] = bkg_data[i]
        coefs[7] = w[i, -1, 3]

        if nlayers:
            for j in range(nlayers):
                coefs[8 + 4*j] = w[i, j + 1, 0]
                coefs[9 + 4*j] = w[i, j + 1, 1]
                coefs[10 + 4*j] = w[i, j + 1, 2]
                coefs[11 + 4*j] = w[i, j + 1, 3]

        abeles_wrapper(
            4*nlayers + 8,
            coefs,
            npoints,
            &y_out_data[npoints * i],
            x_data
        )
    free(coefs_arr)

    return yout



from multiprocessing import Pool, get_context
import warnings as _warnings
import os as _os
import sys as _sys
import functools
from tempfile import mkdtemp
from contextlib import contextmanager
from inspect import getfullargspec as _getargspecf


def preserve_cwd(function):
    """
    Ensures that the original working directory is kept when exiting a function

    Parameters
    ----------
    function : callable

    """

    @functools.wraps(function)
    def decorator(*args, **kwargs):
        cwd = _os.getcwd()
        try:
            return function(*args, **kwargs)
        finally:
            _os.chdir(cwd)

    return decorator


class TemporaryDirectory:
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        # Handle mkdtemp raising an exception
        self.name = None
        self.name = mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                print(
                    f"ERROR: {ex!r} while cleaning up {self!r}",
                    file=_sys.stderr,
                )
                return
            self._closed = True
            if _warn:
                # ResourceWarning
                self._warn(f"ResourceWarning: Implicitly cleaning up {self!r}")

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass


def flatten(seq):
    """
    Flatten a nested sequence.

    Parameters
    ----------
    seq : sequence
        The sequence to flatten

    Returns
    -------
    el : generator
        yields flattened sequences from seq
    """
    for el in seq:
        try:
            iter(el)
            if isinstance(el, (str, bytes)):
                raise TypeError
            yield from flatten(el)
        except TypeError:
            yield el


def unique(seq, idfun=id):
    """
    List of unique values in sequence (by object id, not by value).
    Ordering is preserved.

    Parameters
    ----------
    seq : sequence
    idfun : callable

    Returns
    -------
    p : generator
        yields unique values from l

    Notes
    -----
    Because this function works on object id (by default), it won't work
    for looking at the number of unique values in a numpy array - all
    the entries have different object id's.
    """
    seen = {}
    for item in seq:
        marker = idfun(item)
        if marker not in seen:
            seen[marker] = 1
            yield item


@contextmanager
def possibly_open_file(f, mode="r"):
    """
    Context manager for files.

    Parameters
    ----------
    f : {file-like, Path, str}
        If `f` is a file, then yield the file. If `f` is a str or Path then
        open the file and yield the newly opened file.
        On leaving this context manager the file is closed, if it was opened
        by this context manager (i.e. `f` was a str or Path).
    mode : str, optional
        mode is an optional string that specifies the mode in which the file
        is opened.

    Yields
    ------
    g : file-like
        On leaving the context manager the file is closed, if it was opened by
        this context manager.
    """
    close_file = False
    if (hasattr(f, "read") and hasattr(f, "write")) or f is None:
        g = f
    else:
        g = open(f, mode)
        close_file = True
    yield g
    if close_file:
        g.close()


class MapWrapper:
    """
    Parallelisation wrapper for working with map-like callables, such as
    `multiprocessing.Pool.map`.

    Parameters
    ----------
    pool : int or map-like callable
        If `pool` is an integer, then it specifies the number of threads to
        use for parallelization. If ``int(pool) == 1``, then no parallel
        processing is used and the map builtin is used.
        If ``pool == -1``, then the pool will utilise all available CPUs.
        If `pool` is a map-like callable that follows the same
        calling sequence as the built-in map function, then this callable is
        used for parallelisation.
    context : None, {'spawn', 'fork', 'forkserver'}

    """

    def __init__(self, pool=-1, context=None):
        self.pool = None
        self._mapfunc = map
        self._own_pool = False

        ctx = get_context(context)

        if callable(pool):
            self.pool = pool
            self._mapfunc = self.pool
        else:
            # user supplies a number
            if int(pool) == -1:
                # use as many processors as possible
                self.pool = ctx.Pool()
                self._mapfunc = self.pool.map
                self._own_pool = True
            elif int(pool) in [0, 1]:
                pass
            elif int(pool) > 1:
                # use the number of processors requested
                self.pool = ctx.Pool(processes=int(pool))
                self._mapfunc = self.pool.map
                self._own_pool = True

    def __enter__(self):
        return self

    def __del__(self):
        self.close()
        self.terminate()

    def terminate(self):
        if self._own_pool:
            self.pool.terminate()

    def join(self):
        if self._own_pool:
            self.pool.join()

    def close(self):
        if self._own_pool:
            self.pool.close()

    def __exit__(self, exc_type, exc_value, traceback):
        if self._own_pool:
            self.pool.close()
            self.pool.terminate()

    def __call__(self, func, iterable):
        # only accept one iterable because that's all Pool.map accepts
        try:
            return self._mapfunc(func, iterable)
        except TypeError:
            # wrong number of arguments
            raise TypeError(
                "The map-like callable must be of the"
                " form f(func, iterable)"
            )

    def map(self, func, iterable):
        return self(func, iterable)


def getargspec(f):
    return _getargspecf(f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    division,
    print_function,
    absolute_import,
    unicode_literals,
)

from .sampler import *
from .interruptible_pool import InterruptiblePool
from .mpi_pool import MPIPool
from . import util

__version__ = "1.0.0"



"""
Deals with GUI aspects of MCMC
"""

from importlib import resources
from pathlib import Path

import numpy as np
from qtpy import QtCore, QtGui, QtWidgets, uic
from qtpy.compat import getopenfilename

from refnx.analysis import (
    load_chain,
    process_chain,
    autocorrelation_chain,
    integrated_time,
    GlobalObjective,
    Objective,
)
import refnx.reflect._app
from refnx.reflect import Structure
from matplotlib.backends.backend_qtagg import (
    FigureCanvasQTAgg as FigureCanvas,
)
from matplotlib.figure import Figure


UI_LOCATION = resources.files(refnx.reflect._app) / "ui"


class SampleMCMCDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = uic.loadUi(UI_LOCATION / "sample_mcmc.ui", self)


class ProcessMCMCDialog(QtWidgets.QDialog):
    def __init__(self, objective, chain, folder=None, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = uic.loadUi(UI_LOCATION / "process_mcmc.ui", self)

        self.objective = objective
        self.folder = folder
        self.chain = chain

        if folder is None:
            self.folder = Path.cwd()

        if self.chain is None:
            model_file_name, ok = getopenfilename(self, "Select chain file")
            if not ok:
                return
            self.folder = Path(model_file_name).parent
            try:
                self.chain = load_chain(model_file_name)
            except Exception as e:
                # a chain load will go wrong quite often I'd expect
                self.chain = None
                print(repr(e))
                return

        if len(self.chain.shape) == 3:
            steps, walkers, varys = self.chain.shape
            self.chain_size.setText(
                f"steps: {steps}, walkers: {walkers}, varys: {varys}"
            )
        else:
            steps, temps, walkers, varys = self.chain.shape
            self.chain_size.setText(
                f"steps: {steps}, temps: {temps}, "
                f"walkers: {walkers}, varys: {varys}"
            )

        self.total_samples.setText(f"Total samples: {steps * walkers}")

        self.burn.setMaximum(steps - 1)
        self.thin.setMaximum(steps - 1)

        acfs = autocorrelation_chain(self.chain)
        time = integrated_time(acfs, tol=1, quiet=True)
        self.autocorrelation_time.setText(
            f"Estimated Autocorrelation Time: {time}"
        )

    @QtCore.Slot(int)
    def on_burn_valueChanged(self, val):
        self.recalculate()

    @QtCore.Slot(int)
    def on_thin_valueChanged(self, val):
        self.recalculate()

    def recalculate(self):
        nthin = self.thin.value()
        nburn = self.burn.value()

        lchain = self.chain[nburn::nthin]

        if len(lchain.shape) == 3:
            steps, walkers, varys = lchain.shape
        else:
            steps, temps, walkers, varys = lchain.shape

        self.total_samples.setText(f"Total samples: {steps * walkers}")

        acfs = autocorrelation_chain(lchain)
        time = integrated_time(acfs, tol=1, quiet=True)
        self.autocorrelation_time.setText(
            f"Estimated Autocorrelation Time: {time}"
        )
        self.nplot.setMaximum(steps * walkers)

    @QtCore.Slot()
    def on_buttonBox_accepted(self):
        nthin = self.thin.value()
        nburn = self.burn.value()
        _process_chain(
            self.objective, self.chain, nburn, nthin, folder=self.folder
        )


def _process_chain(objective, chain, nburn, nthin, folder=None):
    # processes the chain for the ProcessMCMCDialog
    if folder is None:
        folder = Path.cwd()

    process_chain(objective, chain, nburn=nburn, nthin=nthin)

    # plot the Autocorrelation function of the chain
    acfs = autocorrelation_chain(chain, nburn=nburn, nthin=nthin)

    fig = Figure()
    FigureCanvas(fig)

    ax = fig.add_subplot(111)
    ax.plot(acfs)
    ax.set_ylabel("autocorrelation")
    ax.set_xlabel("step")
    fig.savefig(folder / "steps-autocorrelation.png")


def _plots(obj, nplot=0, folder=None):
    # create graphs of reflectivity and SLD profiles
    # steps.png, steps_corner.png, steps_sld.png, steps_param_{i}.png
    # Also writes out arrays:
    #      steps.npy, steps_sld.npy
    if folder is None:
        folder = Path.cwd()

    fig = Figure()
    FigureCanvas(fig)

    _, ax = obj.plot(samples=nplot, fig=fig)
    ax.set_ylabel("R")
    ax.set_xlabel("Q / $\\AA$")
    fig.savefig(folder / "steps.png", dpi=1000)

    rng = np.random.default_rng(1)

    # write out a file with the reflectivities
    g = obj._generate_generative_mcmc(ngen=nplot, random_state=rng)
    arr = np.vstack([a for a in g])
    np.save(folder / "steps.npy", arr)

    # corner plot
    try:
        import corner

        kwds = {}
        var_pars = obj.varying_parameters()
        chain = np.array([par.chain for par in var_pars])
        labels = [par.name for par in var_pars]
        chain = chain.reshape(len(chain), -1).T
        kwds["labels"] = labels
        kwds["quantiles"] = [0.16, 0.5, 0.84]
        fig2 = corner.corner(chain, **kwds)

        fig2.savefig(folder / "steps_corner.png")
    except ImportError:
        pass

    # plot sld profiles
    if isinstance(obj, GlobalObjective):
        fig3 = Figure()
        FigureCanvas(fig3)
        ax3 = fig3.add_subplot(111)

        if nplot > 0:
            for idx, o in enumerate(obj.objectives):
                if hasattr(o.model, "structure"):
                    o.model.structure.plot(samples=nplot, fig=fig3)

                # write out all the data for the SLD profiles
                rng = np.random.default_rng(1)
                g = o.model.structure._generate_sld_profile_mcmc(
                    samples=nplot, random_state=rng
                )
                arr = np.vstack([a for a in g])
                np.save(folder / f"steps_sld_{idx}.npy", arr)

        for o in obj.objectives:
            if hasattr(o.model, "structure"):
                ax3.plot(*o.model.structure.sld_profile(), zorder=20)

        ax3.set_ylabel("SLD / $10^{-6}\\AA^{-2}$")
        ax3.set_xlabel("z / $\\AA$")

    elif isinstance(obj, Objective) and hasattr(obj.model, "structure"):
        fig3 = Figure()
        FigureCanvas(fig3)
        fig3, ax3 = obj.model.structure.plot(samples=nplot, fig=fig3)

        # write out all the data for the SLD profiles
        rng = np.random.default_rng(1)
        g = obj.model.structure._generate_sld_profile_mcmc(
            samples=nplot, random_state=rng
        )
        arr = np.vstack([a for a in g])
        np.save(folder / "steps_sld.npy", arr)

    fig3.savefig(folder / "steps_sld.png", dpi=1000)

    # plot the chains so one can see when parameters reach
    # 'equilibrium values'
    for i, vp in enumerate(obj.varying_parameters()):
        label = vp.name
        fig = Figure()
        FigureCanvas(fig)
        ax = fig.add_subplot(111)
        ax.set_ylabel(label)

        for j in range(100):
            ax.plot(vp.chain[:, j].flat)

        fig.savefig(folder / f"steps_param_{i}.png")



import numpy as np
import matplotlib.pyplot as plt
import matplotlib

from refnx.dataset import ReflectDataset
from refnx.analysis import CurveFitter, Objective, Transform
from refnx.reflect import ReflectModel, SLD

matplotlib.pyplot.rcParams['figure.figsize'] = (10.0, 10.0)
matplotlib.pyplot.rcParams['figure.dpi'] = 600


DATASET_NAME = 'c_PLP0011859_q.txt'

# load the data
data = ReflectDataset(DATASET_NAME)

si = SLD(2.07, name='Si')
sio2 = SLD(3.47, name='SiO2')
film = SLD(2, name='film')
d2o = SLD(6.36, name='d2o')

structure = si | sio2(30, 3) | film(250, 3) | d2o(0, 3)
structure[1].thick.setp(vary=True, bounds=(15., 50.))
structure[1].rough.setp(vary=True, bounds=(1., 6.))
structure[2].thick.setp(vary=True, bounds=(200, 300))
structure[2].sld.real.setp(vary=True, bounds=(0.1, 3))
structure[2].rough.setp(vary=True, bounds=(1, 6))

model = ReflectModel(structure, bkg=9e-6, scale=1.)
model.bkg.setp(vary=True, bounds=(1e-8, 1e-5))
model.scale.setp(vary=True, bounds=(0.9, 1.1))

# fit on a logR scale, but use weighting
objective = Objective(model, data, transform=Transform('logY'),
                      use_weights=True)

# create the fit instance
fitter = CurveFitter(objective)

# do the fit
res = fitter.fit(method='differential_evolution')

# see the fit results
print(objective)

fig = plt.figure()
ax = fig.add_subplot(2, 1, 1)
ax.scatter(data.x, data.y, label=DATASET_NAME)
ax.semilogy()
ax.plot(data.x, model.model(data.x, x_err=data.x_err), label='fit')
plt.xlabel('Q')
plt.ylabel('logR')
plt.legend()
ax2 = fig.add_subplot(2, 1, 2)
z, rho_z = structure.sld_profile()
ax2.plot(z, rho_z)



import numba
import numpy as np
import cmath


def numba_parratt(
    q,
    slabs,
    scale=1.0,
    bkg=0.0,
    threads=numba.config.NUMBA_DEFAULT_NUM_THREADS,
):
    current_threads = numba.get_num_threads()
    if threads == -1:
        threads = numba.config.NUMBA_DEFAULT_NUM_THREADS
    threads = np.clip(threads, 1, numba.config.NUMBA_DEFAULT_NUM_THREADS)
    numba.set_num_threads(threads)

    qvals = np.asarray(q).astype(float, copy=False)
    flatq = np.ravel(qvals)
    reflectivity = _numba_kernel(
        flatq, slabs[:, 0], slabs[:, 1], slabs[:, 2], slabs[:, 3], scale, bkg
    )

    numba.set_num_threads(current_threads)

    return np.reshape(reflectivity, qvals.shape)


@numba.jit(
    numba.float64[:](
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64,
        numba.float64,
    ),
    nopython=True,
    parallel=False,
    cache=True,
)
def _numba_kernel(q, d, sldr, sldi, sigma, scale, bkg):
    nlayers = d.shape[0] - 2
    points = q.shape[0]

    R = np.empty_like(q)

    sld = 1j * (np.abs(sldi) + 1e-30)
    sld += sldr - sldr[0]
    sld[0] = 0.0
    sld *= np.pi * 4e-6

    for qi in numba.prange(points):
        q2 = 0.25 * q[qi] ** 2
        kn = cmath.sqrt(q2 - sld[-2])
        kn_next = cmath.sqrt(q2 - sld[-1])
        # Fresnel reflectivity for the interfaces
        RRJ = RRJ_1 = (
            (kn - kn_next)
            / (kn + kn_next)
            * cmath.exp(-2.0 * kn * kn_next * sigma[-1] ** 2)
        )

        kn_next = kn
        for lj in range(nlayers - 1, -1, -1):
            kn = cmath.sqrt(q2 - sld[lj])
            # Fresnel reflectivity for the interfaces
            rj = (
                (kn - kn_next)
                / (kn + kn_next)
                * cmath.exp(-2.0 * kn * kn_next * sigma[lj + 1] ** 2)
            )
            beta = cmath.exp(-2.0j * kn_next * d[lj + 1])

            RRJ = (rj + RRJ_1 * beta) / (1.0 + RRJ_1 * beta * rj)
            RRJ_1 = RRJ
            kn_next = kn

        R[qi] = bkg + scale * np.abs(RRJ) ** 2
    return R
